// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch.c
// Description : Display a border around the current terminal with
//               auto-adaptation to window size changes
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha        - Creation
//     06-Aug-2019  R. Koucha        - Made the display faster
//                                   - Bug fix: upon unknown command in the
//                                     main loop, the window size at program's
//                                     startup was used to refresh the frame
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=



#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>


// ----------------------------------------------------------------------------
// Name   : winch_display
// Usage  : Display buffer to reduce the number of write calls to display
//          the frame on the screen
// ----------------------------------------------------------------------------
static char winch_display[4096];


// ----------------------------------------------------------------------------
// Name   : winch_offset
// Usage  : Current offset in the displayu buffer
// ----------------------------------------------------------------------------
static int winch_offset;


// ----------------------------------------------------------------------------
// Name   : winch_buffer_flush
// Usage  : Flush the content of the display buffer
// Return : None
// ----------------------------------------------------------------------------
static void winch_buffer_flush(int fd)
{
int rc;

  if (winch_offset > 0)
  {
    rc = write(fd, winch_display, winch_offset);
    assert(rc == winch_offset);
    winch_offset = 0;
  }
 
} // winch_buffer_flush


// ----------------------------------------------------------------------------
// Name   : winch_buffer_write
// Usage  : Write into an internal buffer and flush it to stdout when it
//          is full
// Return : None
// ----------------------------------------------------------------------------
static void winch_buffer_write(int fd, char *data, size_t data_sz)
{
int i, j;

  for (j = 0, i = winch_offset; (j < (int)data_sz) && (i < (int)sizeof(winch_display)); i ++, j ++)
  {
    winch_display[i] = data[j];
  }

  // New offset
  winch_offset = i;

  // If the display buffer is full ==> Flush it !
  if (winch_offset == (int)sizeof(winch_display))
  {
    winch_buffer_flush(fd);
  }

  // If there are more data to write
  if (j < (int)data_sz)
  {
    // Recursive call
    winch_buffer_write(fd, data + j, data_sz - j);
  }

} // winch_buffer_write



// ----------------------------------------------------------------------------
// Name   : display_border
// Usage  : Display a frame around the terminal's window
// Return : None
// ----------------------------------------------------------------------------
static void display_border(
                           int      fd,
			   unsigned short rows,
                           unsigned short columns
                          )
{
unsigned short r, c;
int            rc;
char           banner[25];
static char   *line = (char *)0;
static size_t  line_sz = 0;

  // Adjust the size of the line buffer
  // (+1 to take in account terminating '\n')
  if ((int)line_sz < (columns + 1))
  {
    line_sz = columns + 1;
    line = (char *)realloc(line, line_sz);
    assert(line);
  }

  r = 0;

  // Print first line
  if (rows > 1)
  {
  unsigned short saved_rows = rows;

    // '\n' to go back to beginning of line after last display of 'colsXrows'
    winch_buffer_write(fd, "\n", 1);
    line[0] = '+';
    for (c = 1; c < (columns - 1); c ++)
    {
      line[c] = '-';
    }
    if (columns - 1)
    {
      line[columns - 1] = '+';
    }
    line[columns] = '\n';
    winch_buffer_write(fd, line, columns + 1);

    rows -= 1;

    if (rows >= 2)
    {
      line[0] = '|';
      for (c = 1; c < (columns - 1); c ++)
      {
        line[c] = ' ';
      }
      if (columns - 1)
      {
        line[columns - 1] = '|';
      }
      line[columns] = '\n';

      // Print intermediate lines
      for (r = 0; r < (rows - 2); r ++)
      {
        winch_buffer_write(fd, line, columns + 1);
      } // End for

      rows -= r;
    }

    // Print last line
    if (rows == 2)
    {
      line[0] = '+';
      for (c = 1; c < (columns - 1); c ++)
      {
        line[c] = '-';
      }
      if (columns - 1)
      {
        line[columns - 1] = '+';
      }
      line[columns] = '\n';
      winch_buffer_write(fd, line, columns + 1);

      rows -= 1;
    }
    else
    {
      assert(1 == rows);
    }

    rc = snprintf(banner, sizeof(banner), "%dx%d: ", columns, saved_rows);
  }
  else
  {
    // '\n' to go back to beginning of line after last display of 'colsXrows'
    rc = snprintf(banner, sizeof(banner), "\n%dx%d: ", columns, rows);
  }

  winch_buffer_write(fd, banner, rc);

  winch_buffer_flush(fd);
} // display_border


// ----------------------------------------------------------------------------
// Name   : winch_winsz
// Usage  : Current size of the terminal
// ----------------------------------------------------------------------------
static struct winsize winch_winsz;


// ----------------------------------------------------------------------------
// Name   : sig_handler
// Usage  : Signal handler for SIGWINCH
// Return : None
// ----------------------------------------------------------------------------
static void sig_handler(int sig)
{
int rc;

  if (SIGWINCH == sig)
  {
    rc = ioctl(0, TIOCGWINSZ, &winch_winsz);
    if (rc != 0)
    {
      fprintf(stderr, "ioctl(TIOCGWINSZ): '%m' (%d)\n", errno);
    }
    display_border(1, winch_winsz.ws_row, winch_winsz.ws_col);
  }

} // sig_handler


// ----------------------------------------------------------------------------
// Name   : winch_usage
// Usage  : Program's help
// Return : None
// ----------------------------------------------------------------------------
static void winch_usage(char *pname)
{

  fprintf(stderr,
          "Usage: %s [-i signum] port\n"
          "\n"
          "     -i signum: Ignore signum signal\n"
          "\n" 
          ,
          basename(pname)
         );

} // winch_usage


// ----------------------------------------------------------------------------
// Name   : main
// Usage  : Program's entry point
// Return : 0, if OK
//          !0, if error
// ----------------------------------------------------------------------------
int main(int ac, char *av[])
{
char         line[256];
size_t       l;
int          rc;
extern int   optind, opterr, optopt;
sighandler_t rcsig;
unsigned int ignore_sig_mask = 0;

  // Disable error messages from the library
  opterr = 0;

  while ((rc = getopt(ac, av, ":i:")) >= 0)
  {
    switch(rc)
    {
      case 'i': // Ignore signal
      {
      int sig = atoi(optarg);

        ignore_sig_mask |= (1 << sig);
        rcsig = signal(sig, SIG_IGN);
        if (SIG_ERR == rcsig)
	{
          fprintf(stderr, "signal(%d): '%m' (%d)\n", sig, errno);
          return 1;
	}

        printf("Ignoring signal %d\n", sig);
      }
      break;

      case ':':
      {
        fprintf(stderr, "Missing value for option '%c'!\n", optopt);
        winch_usage(av[0]);
        return 1;
      }
      break;

      case '?':
      {
        fprintf(stderr, "Unknown option '%c'!\n", optopt);
        winch_usage(av[0]);
        return 1;
      }
      break;

      default:
      {
        winch_usage(av[0]);
        return 1;
      }
      break;
    } // End switch

  } // End while

  // Missing port ?
  if (av[optind])
  {
    fprintf(stderr, "Unexpected arguments!\n");
    return 1;
  }

  // First frame with the current size of the window
  rc = ioctl(0, TIOCGWINSZ, &winch_winsz);
  if (rc != 0)
  {
    fprintf(stderr, "ioctl(TIOCGWINSZ): '%m' (%d)\n", errno);
    return 1;
  }
  display_border(1, winch_winsz.ws_row, winch_winsz.ws_col);

  // Capture SIGWINCH
  rcsig = signal(SIGWINCH, sig_handler);
  if (SIG_ERR == rcsig)
  {
    fprintf(stderr, "signal(SIGWINCH): '%m' (%d)\n", errno);
    return 1;
  }

  while (fgets(line, sizeof(line), stdin))
  {
    l = strlen(line);
    if (l > 0)
    {
      if ('\n' == line[l - 1])
      {
        line[l - 1] = '\0';
      }

      if (!strcmp(line, "exit"))
      {
        break;
      }
    } // End if line is not empty

    // Refresh the frame after the input from the operator
    display_border(1, winch_winsz.ws_row, winch_winsz.ws_col);
  } // End while

  printf("Exiting...\n");

  return 0;

} // main
