// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_client.c
// Description : Client's main program
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha           - Creation
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <libgen.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/signalfd.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/ioctl.h>

#include "winch_util.h"
#include "winch_msg.h"




// ----------------------------------------------------------------------------
// Name   : winch_update_winsz
// Usage  : Send a WINCH message to the server to update the window's size
//          on the PTY of the winch's program
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_update_winsz(int sfd)
{
struct winsize winsz;
int            rc;
unsigned char  bytes[4];

  // Get the current window size
  rc = winch_ioctl(0, TIOCGWINSZ, &winsz);
  if (rc != 0)
  {
    WINCH_ERR("ioctl(TIOCGWINSZ): '%m' (%d)\n", errno);
    rc = -1;
    goto err;
  }

  // The message is:
  //    . columns (2 bytes)
  //    . rows (2 bytes)  
  WINCH_MSG_HTON_16(winsz.ws_col, &(bytes[0]));
  WINCH_MSG_HTON_16(winsz.ws_row, &(bytes[2]));

  // Send a WINCH message to the server
  rc = winch_send_msg(sfd, WINCH_MSG_WINSZ, 4, bytes);
  if (rc != 0)
  {
    WINCH_ERR("send(): '%m' (%d)\n", errno);
    rc = -1;
    goto err;
  }

  rc = 0;

err:

  return rc;

} // winch_update_winsz



// ----------------------------------------------------------------------------
// Name   : main
// Usage  : Client's main program
// Return : 0, if OK
//          !0, if error
// ----------------------------------------------------------------------------
int main(int ac, char *av[])
{
struct addrinfo          hints;
struct addrinfo         *result, *rp;
int                      s;
int                      rc, rc1;
sigset_t                 sigset, old_sigset;
int                      sigfd = -1;
int                      sfd = -1;
fd_set                   fdset, sav_fdset;
int                      nfds;
int                      i;
int                      ready;
struct signalfd_siginfo  siginfo;
int                      offset;
char                     buf[128];
sighandler_t             rcsig;

  if (ac != 3)
  {
    WINCH_ERR("Usage: %s host port\n", basename(av[0]));
    return 1;
  }

  // Initialize the utilities
  winch_util_init();

  // Initialize the network service
  winch_msg_init();

  // Ignore SIGPIPE (to get EPIPE when the other side of the socket closes)
  rcsig = signal(SIGPIPE, SIG_IGN);
  if (SIG_ERR == rcsig)
  {
    WINCH_ERR("signal(SIGPIPE): '%m' (%d)\n", errno);
    return 1;
  }

  // Block SIGWINCH
  rc = sigemptyset(&sigset);
  if (rc != 0)
  {
    WINCH_ERR("sigemptyset(): '%m' (%d)\n", errno);
    return 1;
  }

  rc = sigaddset(&sigset, SIGWINCH);
  if (rc != 0)
  {
    WINCH_ERR("sigaddset(SIGWINCH): '%m' (%d)\n", errno);
    return 1;
  }

  // The set of blocked signals is the union of the current set
  // and the "sigset" argument
  rc = sigprocmask(SIG_BLOCK, &sigset, &old_sigset);
  if (rc != 0)
  {
    WINCH_ERR("sigprocmask(): '%m' (%d)\n", errno);
    return 1;
  }

  sigfd = signalfd(-1, &sigset, 0);
  if (rc != 0)
  {
    WINCH_ERR("signalfd(): '%m' (%d)\n", errno);
    return 1;
  }

  WINCH_DISPLAY("Connecting to %s/%s\n", av[1], av[2]);

  // Get the address of the server
  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM; /* TCP socket */
  hints.ai_flags = 0;
  hints.ai_protocol = 0;
  s = getaddrinfo(av[1], av[2], &hints, &result);
  if (s != 0)
  {
    WINCH_ERR("getaddrinfo(%s, %s): Error '%s' (%d)\n", av[1], av[2], gai_strerror(s), s);
    return 1;
  }

  /*
     getaddrinfo() returns a list of address structures.
     Try each address until we successfully connect(2).
     If socket(2) (or connect(2)) fails, we (close the socket
     and) try the next address.
  */
  for (rp = result; rp != NULL; rp = rp->ai_next)
  {
    sfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
    if (sfd == -1)
    {
      continue;
    }

    if (connect(sfd, rp->ai_addr, rp->ai_addrlen) != -1)
    {
      break;
    }

    close(sfd);
  } // End for

  // No address succeeded
  if (rp == NULL)
  {
    WINCH_ERR("Could not connect\n");
    rc = 1;
    goto err;
  }

  freeaddrinfo(result);

  WINCH_DISPLAY("Connected to %s/%s\n", av[1], av[2]);  

  rc = winch_update_winsz(sfd);
  if (0 != rc)
  {
    rc = 1;
    goto err;
  }

  // Mainloop
  FD_ZERO(&sav_fdset);
  FD_SET(0, &sav_fdset);
  FD_SET(sfd, &sav_fdset);
  FD_SET(sigfd, &sav_fdset);
  if (sigfd > sfd)
  {
    nfds = sigfd + 1;
  }
  else
  {
    nfds = sfd + 1;
  }

  offset = 0;
  while (1)
  {
    // Wait for an event
    fdset = sav_fdset;
    rc = winch_select(nfds, &fdset, 0, 0, 0);
    switch(rc)
    {
      case -1:
      {
        WINCH_ERR("select(): '%m' (%d)\n", errno);
        rc = 1;
        goto err;
      }
      break;

      default: // There is an event
      {
        ready = rc;

        // SIGWINCH ?
	if (ready && FD_ISSET(sigfd, &fdset))
	{
          ready --;

          rc = winch_read(sigfd, &siginfo, sizeof(siginfo));
          if (rc < 0)
	  {
            WINCH_ERR("read(): '%m' (%d)\n", errno);
            rc = 1;
            goto err;
	  }

          if (siginfo.ssi_signo == SIGWINCH)
	  {
            rc = winch_update_winsz(sfd);
            if (0 != rc)
            {
              rc = 1;
              goto err;
            }
	  }
          else
	  {
            WINCH_ERR("Unexpected signal %d\n", siginfo.ssi_signo);
            rc = 1;
            goto err;
	  }
	} // End if signal

        // Data from server
        if (ready && FD_ISSET(sfd, &fdset))
        {
          ready --;

          rc = winch_read(sfd, buf, sizeof(buf));
          if (rc > 0)
	  {
            // Forward to stdout
            rc1 = winch_write(1, buf, rc);
            if (rc1 != rc)
            {
              if (rc1 >= 0)
	      {
                errno = EIO;
	      }
              WINCH_ERR("write(stdout): '%m' (%d)\n", errno);
              rc = 1;
              goto err;
	    }
          }
          else if (rc < 0)
          {
            WINCH_ERR("read(%d): '%m' (%d)\n", sfd, errno);
            rc = 1;
            goto err;
          }
          else
	  {
            WINCH_DISPLAY("Broken connection\n");
            goto out;
          }
	} // End if data from server

        // Data from user ?
        if (ready && FD_ISSET(0, &fdset))
	{
          ready --;

          rc = winch_read(0, buf + offset, sizeof(buf) - offset);
          if (rc > 0)
	  {
            for (i = offset; i < (offset + rc); i ++)
	    {
              if ('\n' == buf[i])
	      {
                buf[i] = '\0';

                // Send the command to the remote winch program
                rc = winch_send_msg(sfd, WINCH_MSG_CMD, i + 1, (unsigned char *)buf);
                if (0 != rc)
		{
                  rc = 1;
                  goto err;
		}

                offset = 0;
	      }
	    } // End for
	  }
          else
	  {
            if (0 == rc)
	    {
              // Line too long ?
              errno = ENOSPC;
	    }

            WINCH_ERR("read(stdin): '%m' (%d)\n", errno);
            rc = 1;
            goto err;
	  }
	} // End if data from user
      }
      break;
    } // End switch
  } // End Main Loop

out:

  rc = 0;

err:

  if (sfd >= 0)
  {
    shutdown(sfd, SHUT_RDWR);
    close(sfd);
  }

  if (sigfd >= 0)
  {
    close(sigfd);
  }

  return rc;

} // main
