// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_fsm.c
// Description : Finite State Machine service
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha           - Creation
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=





/*
    Events are managed in a FIFO manner
*/


#include <errno.h>
#include <stdlib.h>
#include <stdio.h>


#include "winch_fsm.h"
#include "winch_util.h"




static winch_link_t winch_events;



// ----------------------------------------------------------------------------
// Name   : winch_new_event
// Usage  : Allocate an event structure
// Return : event, if OK
//          WINCH_EVENT_NULL, if error
// ----------------------------------------------------------------------------
winch_event_t *winch_new_event(void)
{
winch_event_t *e;

  e = (winch_event_t *)malloc(sizeof(winch_event_t));
  if (!e)
  {
    // Errno is set
    return WINCH_EVENT_NULL;
  }

  e->u.data = (void *)0;
  e->free_data = 1;
  e->client = (winch_client_t *)0;

  return e;

} // winch_new_event


// ----------------------------------------------------------------------------
// Name   : winch_renew_event
// Usage  : Reallocate an event structure
// Return : event, if OK
//          WINCH_EVENT_NULL, if error
// ----------------------------------------------------------------------------
winch_event_t *winch_renew_event(winch_event_t *e)
{
  if (!e)
  {
    errno = EINVAL;
    return WINCH_EVENT_NULL;
  }

  if (e->free_data && e->u.data)
  {
    free(e->u.data);
  }

  e->u.data = (void *)0;
  e->free_data = 1;
  e->client = (winch_client_t *)0;

  return e;

} // winch_renew_event


// ----------------------------------------------------------------------------
// Name   : winch_put_event
// Usage  : Link an event in the event list
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_put_event(winch_event_t *e)
{

  if (!e)
  {
    errno = EINVAL;
    return -1;
  }

  winch_link_tail(&(e->link), &winch_events);

  return 0;

} // winch_put_event


// ----------------------------------------------------------------------------
// Name   : winch_get_event
// Usage  : Unlink the first event from the event list
// Return : event, if OK
//          WINCH_EVENT_NULL, if list is empty
// ----------------------------------------------------------------------------
winch_event_t *winch_get_event(void)
{
winch_link_t *p;

  p = winch_events.head;
  if (p)
  {
    winch_unlink(p, &winch_events);
    return WINCH_LINK2EVENT(p);
  }

  return WINCH_EVENT_NULL;
} // winch_get_event



// ----------------------------------------------------------------------------
// Name   : winch_free_event
// Usage  : Free an event
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_free_event(winch_event_t *e)
{

  if (!e)
  {
    errno = EINVAL;
    return -1;
  }

  if (e->free_data && e->u.data)
  {
    free(e->u.data);
    e->u.data = (void *)0;
  }

  // For debug purposes
  e->evt = e->fsm = -1;
  e->client = (winch_client_t *)0;

  free(e);

  return 0;

} // winch_free_event


void winch_fsm_purge_client_events(winch_client_t *client)
{
winch_link_t  *p, *next;
winch_event_t *e;

  p = winch_events.head;
  while (p)
  {
    e = WINCH_LINK2EVENT(p);

    // If this event concerns a client FSM
    if (WINCH_FSM_CLIENT == e->fsm)
    {
      // If it is an event for this client
      if (client == e->client)
      {
        // Save the next event link
        next = p->next;

        WINCH_DBG(3, "Purging pending event %s for client#%d\n", winch_event_id2str(e->evt), client->pid);

        // Unlink the event
        winch_unlink(p, &winch_events);

        // Free the event
        winch_free_event(e);

        p = next;
        continue;
      } // End if it is an event for this client
    } // End if this event concerns a client FSM

    p = p->next;
  } // End while

} // winch_fsm_purge_client_events


unsigned int winch_fsm_purge_server_events(int ev_id)
{
winch_link_t  *p, *next;
winch_event_t *e;
unsigned int   nb = 0;

  p = winch_events.head;
  while (p)
  {
    e = WINCH_LINK2EVENT(p);

    // If this event concerns a server FSM
    if (WINCH_FSM_SERVER == e->fsm)
    {
      // If it is the searched event
      if (ev_id == e->evt)
      {
        // Save the next event link
        next = p->next;

        WINCH_DBG(3, "Purging pending event %s for server\n", winch_event_id2str(e->evt));

        // Unlink the event
        winch_unlink(p, &winch_events);

        // Free the event
        winch_free_event(e);

        nb ++;

        p = next;
        continue;
      } // End if it is the searched event
    } // End if this event concerns a server FSM

    p = p->next;
  } // End while

  return nb;

} // winch_fsm_purge_server_events



// ----------------------------------------------------------------------------
// Name   : winch_fsm_engine
// Usage  : Entry point of the FSM
// Return : 0, if OK
//          -1, if error
//          -2, End of server FSM = end of service
// ----------------------------------------------------------------------------
int winch_fsm_engine(void)
{
winch_event_t *e;
int            state;

  do
  {
    e = winch_get_event();

    // If no more events
    if (!e)
    {
      break;
    }

    // Inject the event in the target FSM
    switch(e->fsm)
    {
      case WINCH_FSM_SERVER:
      {
        state = winch_fsm_server_engine(e);
        if (WINCH_SERVER_STATE_END == state)
	{
          WINCH_DISPLAY("End of service...\n");
          return -2;
	}
      }
      break;

      case WINCH_FSM_CLIENT:
      {
        state = winch_fsm_client_engine(e);
      }
      break;

      default:
      {
        errno = EINVAL;
        return -1;
      }
      break;
    } // End switch

    // Error ?
    if (-1 == state)
    {
      return -1;
    }
  } while(1);

  return 0;
} // winch_fsm_engine



// ----------------------------------------------------------------------------
// Name   : winch_state_id2str
// Usage  : Translate a state id into a string
// Return : state as a string
// ----------------------------------------------------------------------------
const char *winch_state_id2str(int state)
{
const char *states[] = {
  "RUNNING",
  "TERMINATING",
  "KILLING",
  "END",
  "CONNECTED",
  "DCNX_TERMINATING",
  "DCNX_KILLING",
  "TERMINATING",
  "KILLING",
  "END"
};
static char bad_state[30];

  if (state >= 0 && (state < WINCH_NB_STATES))
  {
    return states[state];
  }
  else
  {
    (void)snprintf(bad_state, sizeof(bad_state), "??? (%d)", state);
    return bad_state;
  } // End switch

} // winch_state_id2str


// ----------------------------------------------------------------------------
// Name   : winch_event_id2str
// Usage  : Translate an event id into a string
// Return : event as a string
// ----------------------------------------------------------------------------
const char *winch_event_id2str(int evt)
{
const char *evts[] =
{
  "CNX",
  "SIGCHLD",
  "SIGTERM",
  "SIGKILL",
  "SIGINT",
  "WINCH",
  "TIMEOUT",
  "PTY_ERROR",
  "NET_ERROR",
  "TERMINATE",
  "DCNX"
};
static char bad_evt[30];

  if ((evt >= 0) && (evt < WINCH_NB_EVENTS))
  {
    return evts[evt];
  }
  else
  {
    (void)snprintf(bad_evt, sizeof(bad_evt), "??? (%d)", evt);
    return bad_evt;
  }

} // winch_event_id2str
