// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_fsm.h
// Description : Finite State Machine service
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha           - Creation
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=



#ifndef WINCH_FSM_H
#define WINCH_FSM_H

#include <sys/types.h>
#include <sys/select.h>
#include <stddef.h>
#include <signal.h>

#include "winch_link.h"
#include "winch_util.h"
#include "winch_timer.h"



// ----------------------------------------------------------------------------
// Name   : WINCH_TIME_BTW_TERMKILL
// Usage  : Timer between SIGTERM and SIGKILL
//          Timer between SIGKILL and exit (divided by 2 and at least 1 second)
// ----------------------------------------------------------------------------
#define WINCH_TIME_BTW_TERMKILL 5




// ----------------------------------------------------------------------------
// Name   : WINCH_XXX_STATE_XXX
// Usage  : FSM states
// ----------------------------------------------------------------------------
#define WINCH_SERVER_STATE_RUNNING           0
#define WINCH_SERVER_STATE_TERMINATING       1
#define WINCH_SERVER_STATE_KILLING           2
#define WINCH_SERVER_STATE_END               3
#define WINCH_CLIENT_STATE_CONNECTED         4
#define WINCH_CLIENT_STATE_DCNX_TERMINATING  5
#define WINCH_CLIENT_STATE_DCNX_KILLING      6
#define WINCH_CLIENT_STATE_TERMINATING       7
#define WINCH_CLIENT_STATE_KILLING           8
#define WINCH_CLIENT_STATE_END               9

#define WINCH_NB_STATES                     10

#define WINCH_CURRENT_STATE                 20



typedef struct
{
  int          state;
} winch_fsm_t;

#define WINCH_LINK2FSM(p)  \
         (p ? ((winch_fsm_t *)((char *)(p) - offsetof(winch_fsm_t, link))) : NULL)



typedef struct
{
  int   sd;           // Socket of the network connection
  int   pty_master;   // Master side of the PTY
  int   pty_slave;    // Slave side of the PTY
  pid_t pid;          // Pid of the program

  winch_timer_t *kill_timer;  // Timer used by the client to kill the program

  winch_fsm_t fsm;

  winch_link_t link;
} winch_client_t;

#define WINCH_LINK2CLIENT(p)  \
         (p ? ((winch_client_t *)((char *)(p) - offsetof(winch_client_t, link))) : NULL)



typedef struct winch_event
{
  // Used internally by the FSM when allocating events by reusing
  // an existing one:
  //   . Set to 1 at reusing time (to cancel the event freeing at the end of
  //     the transition
  //   . Set to 0 before the transition to trigger the event freeing by default
  int keep;

  int evt;
#define WINCH_EVENT_CNX                   0
#define WINCH_EVENT_SIGCHLD               1
#define WINCH_EVENT_SIGTERM               2
#define WINCH_EVENT_SIGKILL               3
#define WINCH_EVENT_SIGINT                4
#define WINCH_EVENT_WINCH                 5
#define WINCH_EVENT_TIMEOUT               6
#define WINCH_EVENT_PTY_ERROR             7
#define WINCH_EVENT_NET_ERROR             8
#define WINCH_EVENT_TERMINATE             9
#define WINCH_EVENT_DCNX                 10

#define WINCH_NB_EVENTS                  11

  int fsm;
#define WINCH_FSM_SERVER 0
#define WINCH_FSM_CLIENT 1

  winch_client_t *client;

  union         // Data attached to the event
  {
    void *data;
    int   status;
  } u;

  int   free_data; // 1, free the data at freeing time

  winch_link_t link;

} winch_event_t;

#define WINCH_LINK2EVENT(p)  \
         (p ? ((winch_event_t *)((char *)(p) - offsetof(winch_event_t, link))) : NULL)


// ----------------------------------------------------------------------------
// Name   : WINCH_EVENT_NULL
// Usage  : NULL event
// ----------------------------------------------------------------------------
#define WINCH_EVENT_NULL (winch_event_t *)0


// ----------------------------------------------------------------------------
// Name   : winch_old_sigset
// Usage  : Original signal mask
// ----------------------------------------------------------------------------
extern sigset_t winch_old_sigset;




// ----------------------------------------------------------------------------
// Name   : winch_new_event
// Usage  : Allocate an event structure
// Return : event, if OK
//          WINCH_EVENT_NULL, if error
// ----------------------------------------------------------------------------
extern winch_event_t *winch_new_event(void);

// ----------------------------------------------------------------------------
// Name   : winch_renew_event
// Usage  : Reallocate an event structure
// Return : event, if OK
//          WINCH_EVENT_NULL, if error
// ----------------------------------------------------------------------------
extern winch_event_t *winch_renew_event(winch_event_t *e);


// ----------------------------------------------------------------------------
// Name   : winch_get_event
// Usage  : Unlink the first event from the event list
// Return : event, if OK
//          WINCH_EVENT_NULL, if list is empty
// ----------------------------------------------------------------------------
extern winch_event_t *winch_get_event(void);


// ----------------------------------------------------------------------------
// Name   : winch_put_event
// Usage  : Link an event in the event list
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
extern int winch_put_event(winch_event_t *e);

// ----------------------------------------------------------------------------
// Name   : winch_free_event
// Usage  : Free an event
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
extern int winch_free_event(winch_event_t *e);

void winch_fsm_purge_client_events(winch_client_t *client);
unsigned int winch_fsm_purge_server_events(int ev_id);

// ----------------------------------------------------------------------------
// Name   : winch_fsm_engine
// Usage  : Entry point of the FSM
// Return : 0, if OK
//          -1, if error
//          -2, End of server FSM = end of service
// ----------------------------------------------------------------------------
extern int winch_fsm_engine(void);



extern int winch_fsm_server_init(void);
extern int winch_fsm_server_engine(winch_event_t *e);

// ----------------------------------------------------------------------------
// Name   : winch_remove_fd_from_fdset
// Usage  : Remove a file descriptor from the listening set
// Return : None
// ----------------------------------------------------------------------------
extern void winch_remove_fd_from_fdset(
                                int fd
				);

// ----------------------------------------------------------------------------
// Name   : winch_add_fd_in_fdset
// Usage  : Add a file descriptor in the listening set
// Return : None
// ----------------------------------------------------------------------------
extern void winch_add_fd_in_fdset(
                           int fd
			   );


extern const char *winch_cfg_fname;

extern fd_set winch_fdset;

extern int    winch_nfds;

extern int winch_sfd;

extern int winch_fsm_client_init(void);
extern int winch_fsm_client_engine(winch_event_t *e);
extern int winch_fsm_client_poll(fd_set *fdset);
extern winch_client_t *winch_fsm_new_client(int sd);
extern winch_client_t *winch_fsm_pid2client(pid_t pid);
extern winch_client_t *winch_fsm_sd2client(int sd);
extern int winch_fsm_client_killall(int sig);
extern int winch_fsm_client_running(void);

// ----------------------------------------------------------------------------
// Name   : winch_state_id2str
// Usage  : Translate a state id into a string
// Return : state as a string
// ----------------------------------------------------------------------------
extern const char *winch_state_id2str(int state);


// ----------------------------------------------------------------------------
// Name   : winch_event_id2str
// Usage  : Translate an event id into a string
// Return : event as a string
// ----------------------------------------------------------------------------
extern const char *winch_event_id2str(int evt);



// ----------------------------------------------------------------------------
// Name   : WINCH_DBG_UNEXEVT
// Usage  : Display a debug message for an unexpected event in a given state
// ----------------------------------------------------------------------------
#define WINCH_DBG_UNEXEVT(l, e, s) WINCH_DBG((l), "Unexpected event %s in state %s\n", winch_event_id2str(e), winch_state_id2str(s));


// ----------------------------------------------------------------------------
// Name   : WINCH_DBG_UNEXSTATE
// Usage  : Display a debug message for an unexpected state
// ----------------------------------------------------------------------------
#define WINCH_DBG_UNEXSTATE(l, s) WINCH_DBG((l), "Unexpected state %s\n", winch_state_id2str(s));


#endif // WINCH_FSM_H
