// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_fsm_client.c
// Description : Finite State Machine of the client
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha           - Creation
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#include <assert.h>
#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <string.h>
#include <arpa/inet.h>

#include "winch_fsm.h"
#include "winch_util.h"
#include "winch_msg.h"



#define WINCH_EXEC "./winch"


static winch_link_t winch_clients;

static winch_fsm_t *winch_fsm;


// ----------------------------------------------------------------------------
// Name   : winch_msg_winsz_t
// Usage  : Data of the WINCH_MSG_WINSZ message
//             . nb_columns (2 bytes)
//             . nb_rows (2 bytes)
// ----------------------------------------------------------------------------
#define WINCH_MSG_WINSZ_COL_IDX  0
#define WINCH_MSG_WINSZ_ROW_IDX  2


// ----------------------------------------------------------------------------
// Name   : winch_fsm_free_client
// Usage  : Free a client's context
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static void winch_fsm_close_client_cnx(winch_client_t *client)
{

  // Close PTY with program
  if (client->pty_master >= 0)
  {
    WINCH_DBG(5, "Closing PTY fd %d in client#%d\n", client->pty_master, client->pid);

    // Normally this will trigger a SIGHUP signal on slave side
    close(client->pty_master);
    winch_remove_fd_from_fdset(client->pty_master);
    client->pty_master = -1;
  }

  // Close network connection with client
  if (client->sd >= 0)
  {
    WINCH_DBG(5, "Closing network fd %d in client#%d\n", client->sd, client->pid);

    shutdown(client->sd, SHUT_RDWR);
    close(client->sd);
    winch_remove_fd_from_fdset(client->sd);
    client->sd = -1;
  }

} // winch_fsm_close_client_cnx


// ----------------------------------------------------------------------------
// Name   : winch_fsm_free_client
// Usage  : Free a client's context
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static void winch_fsm_free_client(winch_client_t *client)
{
  // Free all pending events concerning this client
  winch_fsm_purge_client_events(client);

  winch_fsm_close_client_cnx(client);

  assert(client->pid >= 1);
  client->pid = -1;

  winch_unlink(&(client->link), &winch_clients);

  free(client);
} // winch_fsm_free_client



// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_connected
// Usage  : CONNECTED state
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_fsm_client_connected(
                                      winch_event_t *e
                                     )
{
winch_client_t *client = e->client;
winch_event_t  *evt;
int             rc;

  switch(e->evt)
  {
    case WINCH_EVENT_WINCH:
    {
    struct winsize winsz;
    unsigned char *data = (unsigned char *)(e->u.data);

      winsz.ws_col = WINCH_MSG_NTOH_16(&(data[WINCH_MSG_WINSZ_COL_IDX]));
      winsz.ws_row = WINCH_MSG_NTOH_16(&(data[WINCH_MSG_WINSZ_ROW_IDX]));

      WINCH_DBG(5, "New window size (%d rows / %d columns) in client#%d\n", winsz.ws_row, winsz.ws_col, client->pid);

      // Update the PTY size ==> This will raise the SIGWINCH signal
      // on the slave side. The child's process will update its display
      // accordingly as it captures SIGWINCH signal.
      rc = winch_ioctl(client->pty_master, TIOCSWINSZ, &winsz);
      if (rc != 0)
      {
        WINCH_ERR("ioctl(%d, TIOCSWINSZ) in client#%d: '%m' (%d)\n", client->pty_master, client->pid, errno);
      }

      return WINCH_CLIENT_STATE_CONNECTED;
    }
    break;

    case WINCH_EVENT_DCNX: // Disconnection from client
    {
      // Close client's connections
      winch_fsm_close_client_cnx(client);

      // Send SIGTERM to client's process
      rc = kill(client->pid, SIGTERM);
      if (0 != rc)
      {
        WINCH_ERR("kill(%d, SIGTERM): '%m' (%d)\n", client->pid, errno);
      }

      // Arm a timer
      evt = winch_renew_event(e);
      if (!evt)
      {
        return -1;
      }
      evt->keep = 1;
      evt->evt = WINCH_EVENT_TIMEOUT;
      evt->fsm = WINCH_FSM_CLIENT;
      evt->client = client;
      evt->u.data = 0;
      evt->free_data = 0;
      
      client->kill_timer = winch_timer_start("term", WINCH_TIME_BTW_TERMKILL, (void *)evt);
      if (!client->kill_timer)
      {
        return -1;
      }

      return WINCH_CLIENT_STATE_DCNX_TERMINATING;  
    }
    break;

    case WINCH_EVENT_PTY_ERROR: // Error on the child's PTY
    case WINCH_EVENT_NET_ERROR: // Error on the client's connection
    {
      // Close client's connections
      winch_fsm_close_client_cnx(client);

      // Send SIGTERM to child program
      rc = kill(client->pid, SIGTERM);
      if (0 != rc)
      {
        WINCH_ERR("kill(%d, SIGTERM): '%m' (%d)\n", client->pid, errno);
      }

      // Arm a timer
      evt = winch_renew_event(e);
      if (!evt)
      {
        return -1;
      }
      evt->keep = 1;
      evt->evt = WINCH_EVENT_TIMEOUT;
      evt->fsm = WINCH_FSM_CLIENT;
      evt->client = client;
      evt->u.data = 0;
      evt->free_data = 0;
      
      client->kill_timer = winch_timer_start("term", WINCH_TIME_BTW_TERMKILL, (void *)evt);
      if (!client->kill_timer)
      {
        return -1;
      }

      return WINCH_CLIENT_STATE_DCNX_TERMINATING;  
    }
    break;

    case WINCH_EVENT_SIGCHLD: // Program (winch) is dead
    {
      // Close client's connections
      winch_fsm_close_client_cnx(client);

      // Send event to the server
      evt = winch_renew_event(e);
      if (!evt)
      {
        return -1;
      }
      evt->keep = 1;
      evt->evt = WINCH_EVENT_SIGCHLD;
      evt->fsm = WINCH_FSM_SERVER;
      evt->u.status = e->u.status;
      evt->free_data = 0;

      rc = winch_put_event(evt);
      if (rc != 0)
      {
        winch_free_event(evt);
        return -1;
      }

      return WINCH_CLIENT_STATE_END;
    }
    break;

    case WINCH_EVENT_SIGTERM:
    {
      // Close client's connections
      winch_fsm_close_client_cnx(client);

      // Send SIGTERM to client's process
      rc = kill(client->pid, SIGTERM);
      if (0 != rc)
      {
        WINCH_ERR("kill(%d, SIGTERM): '%m' (%d)\n", client->pid, errno);
      }

      WINCH_DBG(4, "SIGTERM to process#%d\n", client->pid);

      return WINCH_CLIENT_STATE_TERMINATING;
    }
    break;

    case WINCH_EVENT_SIGKILL:
    {
      // Close client's connections
      winch_fsm_close_client_cnx(client);

      // Send SIGKILL to program
      rc = kill(client->pid, SIGKILL);
      if (0 != rc)
      {
        WINCH_ERR("kill(%d, SIGKILL): '%m' (%d)\n", client->pid, errno);
      }

      WINCH_DBG(4, "SIGKILL to process#%d\n", client->pid);

      return WINCH_CLIENT_STATE_KILLING;
    }
    break;

    default :
    {
      WINCH_DBG_UNEXEVT(3, e->evt, client->fsm.state);
      return -1;
    }
    break;
  } // End switch
} // winch_fsm_client_connected



// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_dcnx_terminating
// Usage  : DCNX_TERMINATING state
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_fsm_client_dcnx_terminating(
                                             winch_event_t *e
                                            )
{
winch_client_t *client = e->client;
winch_event_t  *evt;
int             rc;

  switch(e->evt)
  {
    case WINCH_EVENT_SIGCHLD: // Program (winch) is dead
    {
      // Stop the timer
      assert(client->kill_timer);
      winch_timer_stop(client->kill_timer);
      winch_timer_free(client->kill_timer);
      client->kill_timer = WINCH_TIMER_NULL;

      // Send event to the server
      evt = winch_renew_event(e);
      if (!evt)
      {
        return -1;
      }
      evt->keep = 1;
      evt->evt = WINCH_EVENT_SIGCHLD;
      evt->fsm = WINCH_FSM_SERVER;
      evt->u.status = e->u.status;
      evt->free_data = 0;

      rc = winch_put_event(evt);
      if (rc != 0)
      {
        winch_free_event(evt);
        return -1;
      }

      return WINCH_CLIENT_STATE_END;
    }
    break;

    case WINCH_EVENT_TIMEOUT:
    {
      // Free the timer
      assert(client->kill_timer);
      winch_timer_free(client->kill_timer);
      client->kill_timer = WINCH_TIMER_NULL;

      // Send SIGKILL to program
      rc = kill(client->pid, SIGKILL);
      if (0 != rc)
      {
        WINCH_ERR("kill(%d, SIGKILL): '%m' (%d)\n", client->pid, errno);
      }

      // Arm a timer
      evt = winch_renew_event(e);
      if (!evt)
      {
        return -1;
      }
      evt->keep = 1;
      evt->evt = WINCH_EVENT_TIMEOUT;
      evt->fsm = WINCH_FSM_CLIENT;
      evt->client = client;
      evt->u.data = 0;
      evt->free_data = 0;

      client->kill_timer = winch_timer_start("kill", (WINCH_TIME_BTW_TERMKILL / 2 ?:1), (void *)evt);
      if (!(client->kill_timer))
      {
        return -1;
      }

      return WINCH_CLIENT_STATE_DCNX_KILLING;
    }
    break;

    case WINCH_EVENT_SIGTERM: // SIGTERM from SERVER FSM
    {
      // Stop and free the timer
      assert(client->kill_timer);
      winch_timer_stop(client->kill_timer);
      winch_timer_free(client->kill_timer);
      client->kill_timer = WINCH_TIMER_NULL;

      return WINCH_CLIENT_STATE_DCNX_TERMINATING;
    }
    break;

    case WINCH_EVENT_SIGKILL: // SIGKILL from SERVER FSM
    {
      // Stop and free the timer
      assert(client->kill_timer);
      winch_timer_stop(client->kill_timer);
      winch_timer_free(client->kill_timer);
      client->kill_timer = WINCH_TIMER_NULL;

      // Send SIGKILL to client
      rc = kill(client->pid, SIGKILL);
      if (0 != rc)
      {
        WINCH_ERR("kill(%d, SIGKILL): '%m' (%d)\n", client->pid, errno);
      }

      WINCH_DBG(4, "SIGKILL to process#%d\n", client->pid);

      return WINCH_CLIENT_STATE_KILLING;
    }
    break;

    default :
    {
      WINCH_DBG_UNEXEVT(3, e->evt, client->fsm.state);
      return -1;
    }
    break;
  } // End switch
} // winch_fsm_client_dcnx_terminating




// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_dcnx_killing
// Usage  : DCNX_TERMINATING state
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_fsm_client_dcnx_killing(
                                         winch_event_t *e
                                        )
{
winch_client_t *client = e->client;
winch_event_t  *evt;
int             rc;

  switch(e->evt)
  {
    case WINCH_EVENT_SIGCHLD: // Program (winch) is dead
    {
      // Stop the timer
      assert(client->kill_timer);
      winch_timer_stop(client->kill_timer);
      winch_timer_free(client->kill_timer);
      client->kill_timer = WINCH_TIMER_NULL;

      // Send event to the server
      evt = winch_renew_event(e);
      if (!evt)
      {
        return -1;
      }
      evt->keep = 1;
      evt->evt = WINCH_EVENT_SIGCHLD;
      evt->fsm = WINCH_FSM_SERVER;
      evt->u.status = e->u.status;
      evt->free_data = 0;

      rc = winch_put_event(evt);
      if (rc != 0)
      {
        winch_free_event(evt);
        return -1;
      }

      return WINCH_CLIENT_STATE_END;
    }
    break;

    case WINCH_EVENT_SIGTERM: // SIGTERM from SERVER FSM
    {
      // Stop the timer
      assert(client->kill_timer);
      winch_timer_stop(client->kill_timer);
      winch_timer_free(client->kill_timer);
      client->kill_timer = WINCH_TIMER_NULL;

      // No SIGTERM to send as a SIGKILL is on track

      return WINCH_CLIENT_STATE_TERMINATING;
    }
    break;

    case WINCH_EVENT_SIGKILL: // SIGKILL from SERVER FSM
    {
      // Stop the timer
      assert(client->kill_timer);
      winch_timer_stop(client->kill_timer);
      winch_timer_free(client->kill_timer);
      client->kill_timer = WINCH_TIMER_NULL;

      // No SIGKILL to send as a SIGKILL is on track

      return WINCH_CLIENT_STATE_KILLING;
    }
    break;

    case WINCH_EVENT_TIMEOUT:
    {
      // Free the timer
      assert(client->kill_timer);
      winch_timer_free(client->kill_timer);
      client->kill_timer = WINCH_TIMER_NULL;

      WINCH_DBG(4, "Child %d doesn't want to die ?!\n", client->pid);

      return -1;
    }
    break;

    default :
    {
      WINCH_DBG_UNEXEVT(3, e->evt, client->fsm.state);
      return -1;
    }
    break;
  } // End switch
} // winch_fsm_client_dcnx_killing


// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_terminating
// Usage  : TERMINATING state
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_fsm_client_terminating(
                                        winch_event_t *e
                                       )
{
winch_client_t *client = e->client;
winch_event_t  *evt;
int             rc;

  switch(e->evt)
  {
    case WINCH_EVENT_SIGCHLD:
    {
      // Send event to the server
      evt = winch_renew_event(e);
      if (!evt)
      {
        return -1;
      }
      evt->keep = 1;
      evt->evt = WINCH_EVENT_SIGCHLD;
      evt->fsm = WINCH_FSM_SERVER;
      evt->u.status = e->u.status;
      evt->free_data = 0;

      rc = winch_put_event(evt);
      if (rc != 0)
      {
        winch_free_event(evt);
        return -1;
      }

      return WINCH_CLIENT_STATE_END;
    }
    break;

    case WINCH_EVENT_SIGKILL:
    {
      // Kill the process
      rc = kill(client->pid, SIGKILL);
      if (0 != rc)
      {
        WINCH_ERR("kill(%d, SIGKILL): '%m' (%d)\n", client->pid, errno);
      }

      WINCH_DBG(4, "SIGKILL to process#%d\n", client->pid);

      return WINCH_CLIENT_STATE_KILLING;
    }
    break;

    default :
    {
      WINCH_DBG_UNEXEVT(3, e->evt, client->fsm.state);
      return -1;
    }
    break;
  } // End switch

} // winch_fsm_client_terminating


// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_killing
// Usage  : KILLING state
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_fsm_client_killing(
                                    winch_event_t *e
                                   )
{
winch_client_t *client = e->client;
winch_event_t  *evt;
int             rc;

  switch(e->evt)
  {
    case WINCH_EVENT_SIGCHLD:
    {
      // Send event to the server
      evt = winch_renew_event(e);
      if (!evt)
      {
        return -1;
      }
      evt->keep = 1;
      evt->evt = WINCH_EVENT_SIGCHLD;
      evt->fsm = WINCH_FSM_SERVER;
      evt->u.status = e->u.status;
      evt->free_data = 0;

      rc = winch_put_event(evt);
      if (rc != 0)
      {
        winch_free_event(evt);
        return -1;
      }

      return WINCH_CLIENT_STATE_END;
    }
    break;

    default :
    {
      WINCH_DBG_UNEXEVT(3, e->evt, client->fsm.state);
      return -1;
    }
    break;
  } // End switch

} // winch_fsm_client_killing



// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_engine
// Usage  : Entry point of the client't FSM
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_fsm_client_engine(
                            winch_event_t *e
		           )
{
int             state;
winch_client_t *client;

  // By default we free the event after the transition
  e->keep = 0;

  // Identify the client
  client = e->client;
  winch_fsm = &(client->fsm);

  WINCH_DBG(3, "Client#%d's FSM: state %s <--- event %s\n", client->pid, winch_state_id2str(client->fsm.state), winch_event_id2str(e->evt));

  switch(client->fsm.state)
  {
    case WINCH_CLIENT_STATE_CONNECTED:
    {
      state = winch_fsm_client_connected(e);
    }
    break;

    case WINCH_CLIENT_STATE_DCNX_TERMINATING:
    {
      state = winch_fsm_client_dcnx_terminating(e);
    }
    break;

    case WINCH_CLIENT_STATE_DCNX_KILLING:
    {
      state = winch_fsm_client_dcnx_killing(e);
    }
    break;

    case WINCH_CLIENT_STATE_TERMINATING:
    {
      state = winch_fsm_client_terminating(e);
    }
    break;

    case WINCH_CLIENT_STATE_KILLING:
    {
      state = winch_fsm_client_killing(e);
    }
    break;

    default :
    {
      WINCH_DBG_UNEXSTATE(3, client->fsm.state);
      state = -1;
    }
  } // End switch

  // Free the event unless the transition wanted to keep it
  if (!(e->keep))
  {
    winch_free_event(e);
  }

  if (state != WINCH_CURRENT_STATE)
  {
    winch_fsm->state = state;
  }
  else
  {
    state = winch_fsm->state;
  }

  WINCH_DBG(3, "Client#%d's FSM: New state %s\n", client->pid, winch_state_id2str(state));

  if (WINCH_CLIENT_STATE_END == state)
  {
    // Cleanup the context
    winch_fsm_free_client(client);
  }

  // The client context is freed
  return state;

} // winch_fsm_client_engine


// ----------------------------------------------------------------------------
// Name   : winch_fsm_new_client
// Usage  : Allocate a client's context
// Return : Client's context, if OK
//          0, if error
// ----------------------------------------------------------------------------
winch_client_t *winch_fsm_new_client(int sd)
{
winch_client_t *client;
int             err_sav;
char            name[256];
pid_t           pid;
int             rc;

  client = (winch_client_t *)malloc(sizeof(winch_client_t));
  if (!client)
  {
    // Errno is set
    return (winch_client_t *)0;
  }

  // Create a PTY
  rc = winch_open_pty(&(client->pty_master),
                      &(client->pty_slave),
                      name,
                      sizeof(name)
                     );
  if (0 != rc)
  {
    // Errno is set
    err_sav = errno;
    goto err;
  }

  // Make a child process
  pid = fork();
  switch(pid)
  {
    case -1 : // Error
    {
      err_sav = errno;
      WINCH_ERR("fork(): '%m' (%d)\n", errno);
      goto err;
    }
    break;

    case 0: // Child
    {
    char         *av[10];
    sighandler_t  rcsig;

      // Restablish the default signal mask
      rc = sigprocmask(SIG_SETMASK, &winch_old_sigset, NULL);
      if (rc != 0)
      {
        WINCH_ERR("sigprocmask(): '%m' (%d)\n", errno);
        exit(1);
      }

      // Restablish signal's 
      rcsig = signal(SIGPIPE, SIG_DFL);
      if (SIG_ERR == rcsig)
      {
        WINCH_ERR("signal(SIGPIPE): '%m' (%d)\n", errno);
        exit(1);
      }

      // Close the master side of the PTY
      close(client->pty_master);
      client->pty_master = -1;

      // Redirect input/outputs to PTY
      if (dup2(client->pty_slave, 0) < 0)
      {
        WINCH_ERR("dup2(0): '%m' (%d)\n", errno);
        exit(1);
      }
      if (dup2(client->pty_slave, 1) < 0)
      {
        WINCH_ERR("dup2(1): '%m' (%d)\n", errno);
        exit(1);
      }
      if (dup2(client->pty_slave, 2) < 0)
      {
        WINCH_ERR("dup2(2): '%m' (%d)\n", errno);
        exit(1);
      }

      // Detach the child from the server's terminal
      rc = setsid();
      if (rc != getpid())
      {
        WINCH_ERR("setsid(): '%m' (%d)\n", errno);
        exit(1);
      }

      // Make the PTY the new controlling terminal
      rc = winch_ioctl(0, TIOCSCTTY, 0);
      if (rc != 0)
      {
        WINCH_ERR("ioctl(TIOCSCTTY): '%m' (%d)\n", errno);
        exit(1);
      }

      // Execute winch tool
      av[0] = WINCH_EXEC;
      av[1] = "-i";
      av[2] = "15";
      av[3] = "-i";
      av[4] = "2";
      av[5] = "-i";
      av[6] = "3";
      av[7] = "-i";
      av[8] = "1";
      av[9] = NULL;
      execve(WINCH_EXEC, av, NULL);
      _exit(1);
    }
    break;

    default:
    {
      WINCH_DBG(4, "Forked client#%d with socket %d\n", pid, sd);
      client->pid = pid;

      // Close the slave side of the PTY (used by the child)
      close(client->pty_slave);
      client->pty_slave = -1;

      client->sd = sd;

      // Add the socket to the list of select()
      winch_add_fd_in_fdset(sd);

      // Add the master PTY to the list of select()
      winch_add_fd_in_fdset(client->pty_master);      

      client->fsm.state = WINCH_CLIENT_STATE_CONNECTED;

      client->kill_timer = WINCH_TIMER_NULL;

      WINCH_DBG(3, "Client#%d's FSM: state %s\n", client->pid, winch_state_id2str(client->fsm.state));

      winch_link_head(&(client->link), &winch_clients);
    }
  }// End switch

  return client;

err:

  if (client)
  {
    if (client->pty_master >= 0)
    {
      close(client->pty_master);
      client->pty_master = -1;
    }

    if (client->pty_slave >= 0)
    {
      close(client->pty_slave);
      client->pty_slave = -1;
    }

    free(client);
  } // End if c

  errno = err_sav;

  return (winch_client_t *)0;
} // winch_fsm_new_client


// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_poll
// Usage  : Receive data from client and winch processes
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_fsm_client_poll(fd_set *fdset)
{
winch_client_t            *client;
int                        rc;
WINCH_MSG_SIZE_FIELD_TYPE  data_size;
WINCH_MSG_TYPE_FIELD_TYPE  type;
winch_event_t             *evt;

  client = WINCH_LINK2CLIENT(winch_clients.head);
  while(client)
  {
    // If there are pending data from the client's program
    if ((client->sd > 0) && FD_ISSET(client->sd, fdset))
    {
    char *data;

      // Data from client
      rc = winch_recv_msg(client->sd, &type, &data_size, &data);
      switch(rc)
      {
        case 0:
	{
          switch(type)
	  {
	    case WINCH_MSG_WINSZ:
	    {
	      WINCH_DBG(6, "WINSZ from client#%d\n", client->pid);

              // Send event to the client's FSM
              evt = winch_new_event();
              if (!evt)
              {
                return -1;
              }
              evt->evt = WINCH_EVENT_WINCH;
              evt->fsm = WINCH_FSM_CLIENT;
              evt->client = client;
              evt->u.data = data;
              evt->free_data = 1;

              rc = winch_put_event(evt);
              if (rc != 0)
              {
                winch_free_event(evt);
                return -1;
              }

              // Do not free the data has there are embedded in the previous event
	    }
            break;

	    case WINCH_MSG_CMD:
	    {
	    size_t l;

	      WINCH_DBG(6, "COMMAND from client#%d: '%s'\n", client->pid, data);

              // Replace the terminating NULL by an '\n'
              l = strlen(data);
              data[l] = '\n';

              // Write the command to the PTY to make winch interpret it
              rc = winch_write(client->pty_master, data, l + 1);

              // Free the received message
              free(data);
	    }
            break;

	    default:
	    {
              assert(0);
	    }
            break;
	  } // End switch
	}
        break;

        case -2: // End of connection
	{          
          WINCH_DBG(5, "recv(%d) in client#%d: End of connection\n", client->sd, client->pid);

          // Send event to the client's FSM
          evt = winch_new_event();
          if (!evt)
          {
            return -1;
          }
          evt->evt = WINCH_EVENT_DCNX;
          evt->fsm = WINCH_FSM_CLIENT;
          evt->client = client;
          evt->u.data = NULL;
          evt->free_data = 0;

          rc = winch_put_event(evt);
          if (rc != 0)
          {
            winch_free_event(evt);
            return -1;
          }
	}
        break;

        case -1:
	{
          WINCH_DBG(5, "recv(%d) in client#%d: '%m' (%d)\n", client->sd, client->pid, errno);

          // Send event to the client's FSM
          evt = winch_new_event();
          if (!evt)
          {
            return -1;
          }
          evt->evt = WINCH_EVENT_NET_ERROR;
          evt->fsm = WINCH_FSM_CLIENT;
          evt->client = client;
          evt->u.data = NULL;
          evt->free_data = 0;

          rc = winch_put_event(evt);
          if (rc != 0)
          {
            winch_free_event(evt);
            return -1;
          }
	}
        break;

        default: // Unexpected code
	{
          assert(0);
	}
        break;
      } // End switch
    } // End if data from client

    if ((client->pty_master > 0) && FD_ISSET(client->pty_master, fdset))
    {
    char display[4096];

      // Data from client's program (winch)
      WINCH_DBG(9, "DATA from master PTY#%d of client#%d\n", client->pty_master, client->pid);

      rc = winch_read(client->pty_master, display, sizeof(display));
      if (rc <= 0)
      {
        // When the child terminates, he closes the slave side et this triggers an error
        // here. So, as it is part of the correct behaviour, just display a debug message
        // but not an error
        if (rc < 0)
	{
          WINCH_DBG(5, "read(%d) in client#%d: '%m' (%d)\n", client->pty_master, client->pid, errno);
	}
        else
        {
          WINCH_DBG(5, "PTY#%d closed with client#%d\n", client->pty_master, client->pid);
	}

        // Send event to the client's FSM
        evt = winch_new_event();
        if (!evt)
        {
          return -1;
        }
        evt->evt = WINCH_EVENT_PTY_ERROR;
        evt->fsm = WINCH_FSM_CLIENT;
        evt->client = client;
        evt->u.data = NULL;
        evt->free_data = 0;

        rc = winch_put_event(evt);
        if (rc != 0)
        {
          winch_free_event(evt);
          return -1;
        }
      }
      else
      {
        // Forward the data to the client
        data_size = rc;
        rc = winch_send(client->sd, display, data_size, 0);
        if (rc != (int)data_size)
	{
          WINCH_ERR("winch_send(%d, %zu) to client#%d: '%m' (%d)\n", client->sd, data_size, client->pid, errno);
	}
      }
    } // End if data from winch

    client = WINCH_LINK2CLIENT(client->link.next);
  } // End while

  return 0;
} // winch_fsm_client_poll


// ----------------------------------------------------------------------------
// Name   : winch_fsm_pid2client
// Usage  : Return the client's context corresponding to the pid
// Return : Client's context, if OK
//          0, if error
// ----------------------------------------------------------------------------
winch_client_t *winch_fsm_pid2client(pid_t pid)
{
winch_client_t  *client;

  client = WINCH_LINK2CLIENT(winch_clients.head);
  while(client)
  {
    if (client->pid == pid)
    {
      return client;
    }

    client = WINCH_LINK2CLIENT(client->link.next);
  } // End while

  return (winch_client_t *)0;
} // winch_fsm_pid2client


// ----------------------------------------------------------------------------
// Name   : winch_fsm_sd2client
// Usage  : Return the client's context corresponding to the socket id
// Return : Client's context, if OK
//          0, if error
// ----------------------------------------------------------------------------
winch_client_t *winch_fsm_sd2client(int sd)
{
winch_client_t  *client;

  client = WINCH_LINK2CLIENT(winch_clients.head);
  while(client)
  {
    if (client->sd == sd)
    {
      return client;
    }

    client = WINCH_LINK2CLIENT(client->link.next);
  } // End while

  return (winch_client_t *)0;
} // winch_fsm_sd2client


// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_killall
// Usage  : Send a signal to all the clients
// Return : number of signaled clients, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_fsm_client_killall(int sig)
{
winch_client_t *client;
winch_event_t  *evt;
int             rc;
int             nb;

  nb = 0;
  client = WINCH_LINK2CLIENT(winch_clients.head);
  while(client)
  {
    evt = winch_new_event();
    if (!evt)
    {
      return -1;
    }
    evt->fsm = WINCH_FSM_CLIENT;
    evt->evt = (sig == SIGTERM ? WINCH_EVENT_SIGTERM : WINCH_EVENT_SIGKILL);
    evt->client = client;

    rc = winch_put_event(evt);
    if (0 != rc)
    {
      return -1;
    }

    client = WINCH_LINK2CLIENT(client->link.next);
    nb ++;
  } // End while

  WINCH_DBG(4, "Sent signal '%s' (%d) to %d client%s\n", strsignal(sig), sig, nb, (nb > 1 ? "s" : ""));

  return nb;

} // winch_fsm_client_killall


// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_running
// Usage  : Check if clients are running
// Return : 1, if clients are running
//          0, otherwise
// ----------------------------------------------------------------------------
int winch_fsm_client_running(void)
{
  if (winch_clients.head)
  {
    return 1;
  }
  else
  {
    return 0;
  }
} // winch_fsm_client_running


// ----------------------------------------------------------------------------
// Name   : winch_fsm_client_init
// Usage  : Initialization of the client's FSM
// Return : Client's context, if OK
//          0, if error
// ----------------------------------------------------------------------------
int winch_fsm_client_init(void)
{
struct winsize winsz;

  // Make sure that the serialized fields fit into the host data types
  assert(sizeof(winsz.ws_col) >= sizeof(uint16_t));
  assert(sizeof(winsz.ws_col) >= sizeof(uint16_t));

  return 0;
} // winch_fsm_client_init
