// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_fsm_server.c
// Description : Finite State Machine of the server
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha           - Creation
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#include <errno.h>
#include <sys/types.h>
#include <assert.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>


#include "winch_timer.h"
#include "winch_fsm.h"
#include "winch_util.h"


const char *winch_cfg_fname;

int    winch_sfd = -1;

fd_set winch_fdset;

int    winch_nfds;

unsigned int nb_running_childs;


static winch_fsm_t winch_fsm;

static winch_timer_t *kill_timer;




// ----------------------------------------------------------------------------
// Name   : winch_remove_fd_from_fdset
// Usage  : Remove a file descriptor from the listening set
// Return : None
// ----------------------------------------------------------------------------
void winch_remove_fd_from_fdset(
                                int fd
                               )
{
int i;

  assert(fd >= 0 && FD_ISSET(fd, &winch_fdset));

  FD_CLR(fd, &winch_fdset);

  if (fd == (winch_nfds - 1))
  {
    for (i = fd - 1; i >= 0; i --)
    {
      if (FD_ISSET(i, &winch_fdset))
      {
        winch_nfds = (i + 1);
        break;
      }
    } // end for
  }

  WINCH_DBG(4, "Removed fd#%d from fdset ==> nfds=%d\n", fd, winch_nfds);

} // winch_remove_fd_from_fdset


// ----------------------------------------------------------------------------
// Name   : winch_add_fd_in_fdset
// Usage  : Add a file descriptor in the listening set
// Return : None
// ----------------------------------------------------------------------------
void winch_add_fd_in_fdset(
                           int fd
                          )
{
  assert(fd >= 0 && !FD_ISSET(fd, &winch_fdset));

  FD_SET(fd, &winch_fdset);

  if (fd >= (winch_nfds - 1))
  {
    winch_nfds = (fd + 1);
  }

  WINCH_DBG(4, "Added fd#%d in fdset ==> nfds=%d\n", fd, winch_nfds);

} // winch_add_fd_in_fdset



// ----------------------------------------------------------------------------
// Name   : winch_fsm_server_running
// Usage  : RUNNING state
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_fsm_server_running(
                                    winch_event_t *e
                                   )
{
int             rc;
winch_event_t  *evt;

  switch(e->evt)
  {
    case WINCH_EVENT_CNX:
    {
    socklen_t                addr_len;
    struct sockaddr_in       addr;
    int                      sd;

      // New client

      // Accept the connection
      addr_len = sizeof(addr);
      sd = accept(winch_sfd, (struct sockaddr *)&addr, &addr_len);
      if (sd < 0)
      {
        WINCH_ERR("accept(%d): '%m' (%d)\n", winch_sfd, errno);

        // Go on with the running clients
      }
      else
      {
        WINCH_DBG(4, "Connection request from '%s' (socket %d)\n", inet_ntoa(addr.sin_addr), sd);

        // Allocate a new client
        (void)winch_fsm_new_client(sd);
      }

      return WINCH_SERVER_STATE_RUNNING;
    }
    break;

    case WINCH_EVENT_SIGCHLD:
    {
      // Event coming from the client's FSM. The "status" field is filled
      WINCH_DBG(5, "A client is dead with status 0x%x\n", e->u.status);      

      return WINCH_SERVER_STATE_RUNNING;
    }
    break;

    case WINCH_EVENT_SIGTERM:
    {
      // Terminate all the childs ==> SIGTERM
      if (!winch_fsm_client_running())
      {
        return WINCH_SERVER_STATE_END;
      }

      rc = winch_fsm_client_killall(SIGTERM);
      if (rc < 0)
      {
        return -1;
      }

      // There must be running clients
      assert(rc > 0);

      nb_running_childs = rc;

      evt = winch_renew_event(e);
      if (!evt)
      {
        return -1;
      }
      evt->keep = 1;
      evt->evt = WINCH_EVENT_TIMEOUT;
      evt->fsm = WINCH_FSM_SERVER;

      // Arm a timer
      kill_timer = winch_timer_start("term", WINCH_TIME_BTW_TERMKILL, (void *)evt);
      if (!kill_timer)
      {
        return -1;
      }

      WINCH_DBG(6, "Timer created\n");

      return WINCH_SERVER_STATE_TERMINATING;
    }
    break;

    default :
    {
      WINCH_DBG_UNEXEVT(3,  e->evt, winch_fsm.state);
      return -1;
    }
    break;
  } // End switch


} // winch_fsm_server_running



// ----------------------------------------------------------------------------
// Name   : winch_fsm_server_terminating
// Usage  : TERMINATING state
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_fsm_server_terminating(
                                        winch_event_t *e
                                       )
{
int             rc;
winch_event_t  *evt;

  switch(e->evt)
  {
    case WINCH_EVENT_SIGTERM:
    {
      // Ignore
      return WINCH_CURRENT_STATE;
    }
    break;

    case WINCH_EVENT_CNX:
    {
    socklen_t                addr_len;
    struct sockaddr_in       addr;
    int                      sd;

      // Accept the connection to close it immediately
      addr_len = sizeof(addr);
      sd = accept(winch_sfd, (struct sockaddr *)&addr, &addr_len);
      if (sd < 0)
      {
        WINCH_ERR("accept(%d): '%m' (%d)\n", winch_sfd, errno);

        // Go on with the running clients
      }
      else
      {
        shutdown(sd, SHUT_RDWR);
        close(sd);
      }

      WINCH_DBG(3, "Refused connection request from '%s'\n", inet_ntoa(addr.sin_addr));

      return WINCH_CURRENT_STATE;
    }
    break;

    case WINCH_EVENT_SIGCHLD:
    {
      // Event coming from the client's FSM. The "status" field is filled
      WINCH_DBG(4, "Client is dead with status 0x%x\n", e->u.status);      

      // If no more clients
      assert(nb_running_childs);
      nb_running_childs --;
      if (!nb_running_childs)
      {
        // Stop the timer
        winch_timer_stop(kill_timer);
        winch_timer_free(kill_timer);
        kill_timer = WINCH_TIMER_NULL;

        return WINCH_SERVER_STATE_END;
      }

      return WINCH_CURRENT_STATE;
    }
    break;

    case WINCH_EVENT_TIMEOUT:
    {
      // Free the timer
      assert(kill_timer);
      winch_timer_free(kill_timer);
      kill_timer = WINCH_TIMER_NULL;

      // Kill all the childs
      rc = winch_fsm_client_killall(SIGKILL);
      if (rc < 0)
      {
        return -1;
      }

      // As we are in the event loop, where TIMEOUTs and SIGCHLDs may have been
      // posted, may be all the childs are dead (crossing case)
      if (rc)
      {
        // There are still running childs ==> they will received the SIGKILL

        evt = winch_renew_event(e);
        if (!evt)
        {
          return -1;
        }
        evt->keep = 1;
        evt->evt = WINCH_EVENT_TIMEOUT;
        evt->fsm = WINCH_FSM_SERVER;

        // Arm a timer
        kill_timer = winch_timer_start("kill", (WINCH_TIME_BTW_TERMKILL / 2 ?:1), (void *)evt);
        if (!kill_timer)
        {
          return -1;
        }

        return WINCH_SERVER_STATE_KILLING;
      }
      else // No more childs
      {
      unsigned int nb;

        // This means that the SIGCHLDs and the timeout crossed: the event
        // loop must contain SIGCHLDs
        // Purge all the SIGCHLDs from the event loop
        nb = winch_fsm_purge_server_events(WINCH_EVENT_SIGCHLD);
        assert(nb == nb_running_childs);

        return WINCH_SERVER_STATE_END;
      }
    }
    break;

    default :
    {
      WINCH_DBG_UNEXEVT(3,  e->evt, winch_fsm.state);
      return -1;
    }
    break;
  } // End switch

} // winch_fsm_server_terminating



// ----------------------------------------------------------------------------
// Name   : winch_fsm_server_killing
// Usage  : KILLING state
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_fsm_server_killing(
                                    winch_event_t *e
                                   )
{
  switch(e->evt)
  {
    case WINCH_EVENT_SIGTERM:
    {
      // Ignore
      return WINCH_CURRENT_STATE;
    }
    break;

    case WINCH_EVENT_CNX:
    {
    socklen_t                addr_len;
    struct sockaddr_in       addr;
    int                      sd;

      // Accept the connection to close it immediately
      addr_len = sizeof(addr);
      sd = accept(winch_sfd, (struct sockaddr *)&addr, &addr_len);
      if (sd < 0)
      {
        WINCH_ERR("accept(%d): '%m' (%d)\n", winch_sfd, errno);

        // Go on with the running clients
      }
      else
      {
        shutdown(sd, SHUT_RDWR);
        close(sd);
      }

      WINCH_DBG(3, "Refused connection request from '%s'\n", inet_ntoa(addr.sin_addr));

      return WINCH_CURRENT_STATE;
    }
    break;

    case WINCH_EVENT_SIGCHLD:
    {
      // Event coming from the client's FSM. The "status" field is filled
      WINCH_DBG(4, "Client is dead with status 0x%x\n", e->u.status);      

      // If no more clients
      assert(nb_running_childs);
      nb_running_childs --;
      if (!nb_running_childs)
      {
        // Stop the timer
        winch_timer_stop(kill_timer);
        winch_timer_free(kill_timer);
        kill_timer = WINCH_TIMER_NULL;

        return WINCH_SERVER_STATE_END;
      } // End if no more running childs

      return WINCH_CURRENT_STATE;
    }
    break;

    case WINCH_EVENT_TIMEOUT:
    {
    unsigned int nb;

      // Free the timer
      assert(kill_timer);
      winch_timer_free(kill_timer);
      kill_timer = WINCH_TIMER_NULL;

      // Either childs don't want to die or there is a crossing case where
      // timeout is received before the SIGCHLDs  
      assert(nb_running_childs);

      // Purge the pending SIGCHLDs
      nb = winch_fsm_purge_server_events(WINCH_EVENT_SIGCHLD);
      if (nb != nb_running_childs)
      {
        WINCH_DBG(4, "Some childs don't want to die ?!\n");
      }
      else // All the client died
      {
	// The timeout came before the SIGCHLDs
      } // End if clients are still running

      return WINCH_SERVER_STATE_END;
    }
    break;

    default :
    {
      WINCH_DBG_UNEXEVT(3,  e->evt, winch_fsm.state);
      return -1;
    }
    break;
  } // End switch

} // winch_fsm_server_killing



// ----------------------------------------------------------------------------
// Name   : winch_fsm_server_engine
// Usage  : Entry point of the server's FSM
// Return : new state, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_fsm_server_engine(
                            winch_event_t *e
		           )
{
int state;

  WINCH_DBG(3, "Server's FSM: state %s <--- event %s\n", winch_state_id2str(winch_fsm.state), winch_event_id2str(e->evt));
  
  // By default we free the event after the transition
  e->keep = 0;

  switch(winch_fsm.state)
  {
    case WINCH_SERVER_STATE_RUNNING:
    {
      state = winch_fsm_server_running(e);
    }
    break;

    case WINCH_SERVER_STATE_TERMINATING:
    {
      state = winch_fsm_server_terminating(e);
    }
    break;

    case WINCH_SERVER_STATE_KILLING:
    {
      state = winch_fsm_server_killing(e);
    }
    break;

    default :
    {
      WINCH_DBG_UNEXSTATE(3, winch_fsm.state);
      state = -1;
    }
  } // End switch

  // Free the event unless the transition wanted to keep it
  if (!(e->keep))
  {
    winch_free_event(e);
  }

  if (state != WINCH_CURRENT_STATE)
  {
    winch_fsm.state = state;
  }
  else
  {
    state = winch_fsm.state;
  }

  WINCH_DBG(3, "Server's FSM: New state %s\n", winch_state_id2str(state));

  return state;

} // winch_fsm_server_engine



// ----------------------------------------------------------------------------
// Name   : winch_fsm_server_init
// Usage  : Initialization of the server's FSM
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_fsm_server_init(void)
{

  winch_fsm.state = WINCH_SERVER_STATE_RUNNING;

  return 0;
} // winch_fsm_server_init
