// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_msg.c
// Description : Message management
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     07-Aug-2019  R. Koucha           - Creation
//     11-Aug-2019  R. Koucha           - Added management of endianness
//                                        and CPU dependant data sizes
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=



#include <errno.h>
#include <sys/types.h>
#include <stdlib.h>
#include <inttypes.h>
#include <assert.h>

#include "winch_util.h"
#include "winch_msg.h"





// ----------------------------------------------------------------------------
// Name   : WINCH_MSG_XXX_FIELD_SZ
// Usage  : Size of the message fields on the network
// ----------------------------------------------------------------------------
#define WINCH_MSG_TYPE_FIELD_SZ 1
#define WINCH_MSG_SIZE_FIELD_SZ 4




// ----------------------------------------------------------------------------
// Name   : winch_msg_hdr_t
// Usage  : Network message header
//
//        type
//        size (of the following data)
// ----------------------------------------------------------------------------
typedef unsigned char winch_msg_hdr_t[
                                      WINCH_MSG_TYPE_FIELD_SZ +
                                      WINCH_MSG_SIZE_FIELD_SZ
                                     ];


#define WINCH_MSG_HDR_TYPE_IDX   0
#define WINCH_MSG_HDR_LENGTH_IDX 1



// ----------------------------------------------------------------------------
// Name   : WINCH_MSG_NTOH_16
// Usage  : Translate 2 chars from the network (big endian format) into a
//          16 bits unsigned integer on host 
// Return : None
// ----------------------------------------------------------------------------
uint16_t WINCH_MSG_NTOH_16(unsigned char *bytes)
{
union
{
  uint16_t      v;
  unsigned char bytes[2];
} u;

  u.bytes[0] = bytes[0];
  u.bytes[1] = bytes[1];

  return u.v;
} // WINCH_MSG_NTOH_16



// ----------------------------------------------------------------------------
// Name   : WINCH_MSG_NTOH_32
// Usage  : Translate 4 chars from the network (big endian format) into a
//          32 bits unsigned integer on host 
// Return : None
// ----------------------------------------------------------------------------
uint32_t WINCH_MSG_NTOH_32(unsigned char *bytes)
{
union
{
  uint32_t      v;
  unsigned char bytes[4];
} u;

  u.bytes[0] = bytes[0];
  u.bytes[1] = bytes[1];
  u.bytes[2] = bytes[2];
  u.bytes[3] = bytes[3];

  return u.v;
} // WINCH_MSG_NTOH_32


// ----------------------------------------------------------------------------
// Name   : WINCH_MSG_HTON_16
// Usage  : Translate a 16 bits unsigned int on host into a table of 2 chars
//          to be transfered onto the network (big endian format) 
// Return : None
// ----------------------------------------------------------------------------
void WINCH_MSG_HTON_16(uint16_t v, unsigned char *bytes)
{
union
{
  uint16_t      v;
  unsigned char bytes[2];
} u;

  u.v = v;

  bytes[0] = u.bytes[0];
  bytes[1] = u.bytes[1];

} // WINCH_MSG_HTON_16


// ----------------------------------------------------------------------------
// Name   : WINCH_MSG_HTON_32
// Usage  : Translate a 32 bits unsigned int on host into a table of 4 chars
//          to be transfered onto the network (big endian format) 
// Return : None
// ----------------------------------------------------------------------------
void WINCH_MSG_HTON_32(uint32_t v, unsigned char *bytes)
{
union
{
  uint32_t      v;
  unsigned char bytes[4];
} u;

  u.v = v;

  bytes[0] = u.bytes[0];
  bytes[1] = u.bytes[1];
  bytes[2] = u.bytes[2];
  bytes[3] = u.bytes[3];

} // WINCH_MSG_HTON_32



// ----------------------------------------------------------------------------
// Name   : winch_msg_known_type
// Usage  : Receive a message from the network
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
static int winch_msg_known_type(int type)
{
#define WINCH_MSG_CASE_TYPE(type) case type : return 1

  switch(type)
  {
    WINCH_MSG_CASE_TYPE(WINCH_MSG_WINSZ);
    WINCH_MSG_CASE_TYPE(WINCH_MSG_CMD);
    default: return 0;
  } // End switch

} // winch_msg_known_type


// ----------------------------------------------------------------------------
// Name   : winch_recv_msg
// Usage  : Receive a message from the network
// Return : 0, if OK
//          -1, if error (errno is set)
//          -2, if end of connection
// ----------------------------------------------------------------------------
int winch_recv_msg(
                   int      fd,
                   WINCH_MSG_TYPE_FIELD_TYPE     *type,
                   WINCH_MSG_SIZE_FIELD_TYPE  *data_sz,
                   char   **data    // Dynamically allocated (if *data_sz > 0)
                  )
{
ssize_t          rc;
char            *buffer;
winch_msg_hdr_t  hdr;
int              htype;
size_t           hdata_sz;

  if ((fd < 0) || !type)
  {
    errno = EINVAL;
    return -1;
  }

  // Returned value in case of error
  *type = -1;

  if (!data_sz)
  {
    errno = EINVAL;
    return -1;
  }

  // Returned value in case of error
  *data_sz = 0;

  if (!data)
  {
    errno = EINVAL;
    return -1;
  }

  // Returned value in case of error
  *data = (char *)0;

  // Receive the header
  rc = winch_recv(fd, &hdr, sizeof(hdr), 0);
  if (rc != (ssize_t)sizeof(hdr))
  {
    if (0 == rc)
    {
      return -2;
    }
    else if (rc >= 0)
    {
      errno = EIO;
    }

    return -1;
  } // End if recv() error

  htype = (int)(hdr[WINCH_MSG_HDR_TYPE_IDX]);
  hdata_sz = (size_t)WINCH_MSG_NTOH_32(&(hdr[WINCH_MSG_HDR_LENGTH_IDX]));

  WINCH_DBG(10, "Received header: type=%d/data_sz=%zu\n", htype, hdata_sz);

  // Make sure the we know the type
  if (!winch_msg_known_type(htype))
  {
    errno = EINVAL;
    return -1;
  }

  // Receive the data if any
  if (hdata_sz)
  {
    // Allocate a receiving buffer
    buffer = (char *)malloc(hdata_sz);
    if (!buffer)
    {
      return -1;
    }

    rc = winch_recv(fd, buffer, hdata_sz, 0);
    if (rc != (ssize_t)(hdata_sz))
    {
    int saved_errno;

      if (0 == rc)
      {
        return -2;
      }
      else if (rc >= 0)
      {
        errno = EIO;
      }

      return -1;

      saved_errno = errno;
      free(*data);
      errno = saved_errno;

      return -1;
    }

    *data = buffer;
  } // End if data

  *type = htype;
  *data_sz = hdata_sz;

  return 0;

} // winch_recv_msg



// ----------------------------------------------------------------------------
// Name   : winch_send_msg
// Usage  : Send a message over the network
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_send_msg(
                   int            fd,
                   WINCH_MSG_TYPE_FIELD_TYPE            type,
                   WINCH_MSG_SIZE_FIELD_TYPE         data_sz,
                   unsigned char *data
                  )
{
ssize_t          rc;
winch_msg_hdr_t  hdr;

  if ((fd < 0) || !winch_msg_known_type(type))
  {
    errno = EINVAL;
    return -1;
  }

  if ((data_sz && !data) || (!data_sz && data))
  {
    errno = EINVAL;
    return -1;
  }

  // Send the header
  hdr[WINCH_MSG_HDR_TYPE_IDX] = type;
  WINCH_MSG_HTON_32(data_sz, &(hdr[WINCH_MSG_HDR_LENGTH_IDX]));
  rc = winch_send(fd, &hdr, sizeof(hdr), 0);
  if (rc != (ssize_t)sizeof(hdr))
  {
    if (rc >= 0)
    {
      errno = EIO;
    }

    return -1;
  }

  WINCH_DBG(10, "Sent header: type=%d/data_sz=%zu\n", type, data_sz);

  // Send the data if any
  if (data_sz)
  {
    rc = winch_send(fd, data, data_sz, 0);
    if (rc != (ssize_t)data_sz)
    {
      if (rc >= 0)
      {
        errno = EIO;
      }

      return -1;
    }
  } // End if data

  return 0;
} // winch_send_msg


// ----------------------------------------------------------------------------
// Name   : winch_msg_init
// Usage  : Initialize the network service
// Return : None
// ----------------------------------------------------------------------------
void winch_msg_init(void)
{
  // Make sure that the types on host are equal or bigger than their
  // serialized representation on the network
  assert(sizeof(WINCH_MSG_TYPE_FIELD_TYPE) >= WINCH_MSG_TYPE_FIELD_SZ);
  assert(sizeof(WINCH_MSG_SIZE_FIELD_TYPE) >= WINCH_MSG_SIZE_FIELD_SZ);

} // winch_msg_init
