// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_server.c
// Description : Server main program
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019 R. Koucha           - Creation
//     07-Aug-2019 R. Koucha           - Added options on command line
//                                       (e.g. debug level) 
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


#include <assert.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <libgen.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/select.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/signalfd.h>
#include <sys/wait.h>

#include "winch_util.h"
#include "winch_msg.h"
#include "winch_fsm.h"
#include "winch_timer.h"



// ----------------------------------------------------------------------------
// Name   : winch_old_sigset
// Usage  : Original signal mask
// ----------------------------------------------------------------------------
sigset_t winch_old_sigset;



// ----------------------------------------------------------------------------
// Name   : winch_usage
// Usage  : Program's help
// Return : None
// ----------------------------------------------------------------------------
static void winch_usage(char *pname)
{

  winch_display(2,
                "Usage: %s [-d level] [-c cfg_fname] port\n"
                "\n"
                "     -d level    : Set debug level\n"
                "     -c cfg_fname: Configuration file\n"
                "     port        : TCP port\n"
                "\n" 
                ,
                basename(pname)
               );

} // winch_usage


// ----------------------------------------------------------------------------
// Name   : main
// Usage  : Program's entry point
// Return : 0, if OK
//          !0, if error
// ----------------------------------------------------------------------------
int main(int ac, char *av[])
{
struct addrinfo          hints;
struct addrinfo         *result, *rp;
int                      s;
int                      rc;
fd_set                   fdset;
int                      ready;
struct sockaddr_in      *addrp;
sigset_t                 sigset;
int                      sigfd;
struct signalfd_siginfo  siginfo;
winch_event_t           *event;
extern char             *optarg;
extern int               optind, opterr, optopt;
char                    *port;
sighandler_t             rcsig;

  // Disable error messages from the library
  opterr = 0;

  while ((rc = getopt(ac, av, ":d:c:")) >= 0)
  {
    switch(rc)
    {
      case 'd': // Debug level
      {
        winch_dbg_level = atoi(optarg);
        WINCH_DBG(1, "Debug level: %d\n", winch_dbg_level);
      }
      break;

      case 'c': // Configuration file
      {
        winch_cfg_fname = optarg;
      }
      break;

      case ':':
      {
        winch_display(2, "Missing value for option '%c'!\n", optopt);
        winch_usage(av[0]);
        return 1;
      }
      break;

      case '?':
      {
        winch_display(2, "Unknown option '%c'!\n", optopt);
        winch_usage(av[0]);
        return 1;
      }
      break;

      default:
      {
        winch_usage(av[0]);
        return 1;
      }
      break;
    } // End switch

  } // End while

  // Missing port ?
  if (!(av[optind]))
  {
    winch_display(2, "Missing TCP port!\n");
    winch_usage(av[0]);
    return 1;
  }

  port = av[optind];

  winch_util_init();
  winch_msg_init();

  rc = winch_fsm_server_init();
  if (rc != 0)
  {
    return 1;
  }

  rc = winch_fsm_client_init();
  if (rc != 0)
  {
    return 1;
  }

  // Ignore SIGPIPE (to get EPIPE when the other side of the socket closes)
  rcsig = signal(SIGPIPE, SIG_IGN);
  if (SIG_ERR == rcsig)
  {
    WINCH_ERR("signal(SIGPIPE): '%m' (%d)\n", errno);
    return 1;
  }

  // Capture SIGCHLD, SIGINT, SIGTERM
  rc = sigemptyset(&sigset);
  if (rc != 0)
  {
    WINCH_ERR("sigemptyset(): '%m' (%d)\n", errno);
    return 1;
  }

  rc = sigaddset(&sigset, SIGCHLD);
  if (rc != 0)
  {
    WINCH_ERR("sigaddset(SIGCHLD): '%m' (%d)\n", errno);
    return 1;
  }

  rc = sigaddset(&sigset, SIGTERM);
  if (rc != 0)
  {
    WINCH_ERR("sigaddset(SIGTERM): '%m' (%d)\n", errno);
    return 1;
  }

  rc = sigaddset(&sigset, SIGINT);
  if (rc != 0)
  {
    WINCH_ERR("sigaddset(SIGINT): '%m' (%d)\n", errno);
    return 1;
  }

  rc = sigaddset(&sigset, SIGALRM);
  if (rc != 0)
  {
    WINCH_ERR("sigaddset(SIGALRM): '%m' (%d)\n", errno);
    return 1;
  }

  rc = sigaddset(&sigset, SIGPIPE);
  if (rc != 0)
  {
    WINCH_ERR("sigaddset(SIGPIPE): '%m' (%d)\n", errno);
    return 1;
  }

  // The set of blocked signals is the union of the current set
  // and the "sigset" argument
  rc = sigprocmask(SIG_BLOCK, &sigset, &winch_old_sigset);
  if (rc != 0)
  {
    WINCH_ERR("sigprocmask(): '%m' (%d)\n", errno);
    return 1;
  }

  sigfd = signalfd(-1, &sigset, 0);
  if (rc != 0)
  {
    WINCH_ERR("signalfd(): '%m' (%d)\n", errno);
    return 1;
  }

  // Set the address of the server
  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM; /* TCP socket */
  hints.ai_flags = AI_PASSIVE;
  hints.ai_protocol = 0;
  s = getaddrinfo(NULL, port, &hints, &result);
  if (s != 0)
  {
    WINCH_ERR("getaddrinfo(%s): Error '%s' (%d)\n", port, gai_strerror(s), s);
    return 1;
  }

  /*
     getaddrinfo() returns a list of address structures.
     Try each address until we successfully bind(2).
     If socket(2) (or bind(2)) fails, we (close the socket
     and) try the next address.
  */
  for (rp = result; rp != NULL; rp = rp->ai_next)
  {
    winch_sfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
    if (winch_sfd == -1)
    {
      continue;
    }

    addrp = (struct sockaddr_in *)rp->ai_addr;

    if (bind(winch_sfd, rp->ai_addr, rp->ai_addrlen) == 0)
    {
      break;
    }

    close(winch_sfd);
    winch_sfd = -1;
  } // End for

  // No address succeeded
  if (rp == NULL)
  {
    WINCH_ERR("Could not bind\n");
    rc = 1;
    goto err;
  }

  freeaddrinfo(result);

  {
  int yes = 1;

    rc = setsockopt(winch_sfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
    if (rc == -1)
    {
      WINCH_ERR("setsockopt(SO_REUSEADDR): Error '%m' (%d)\n", errno);
      rc = 1;
      goto err;
    }
  } 

  rc = listen(winch_sfd, 5);
  if (rc == -1)
  {
    WINCH_ERR("listen(%d): Error '%m' (%d)\n", winch_sfd, errno);
    rc = 1;
    goto err;
  }

  WINCH_DBG(1, "Listening on port %s with socket %d at address '%s'\n", port, winch_sfd, inet_ntoa(addrp->sin_addr));

  // Mainloop
  FD_ZERO(&winch_fdset);
  FD_SET(winch_sfd, &winch_fdset);
  FD_SET(sigfd, &winch_fdset);
  if (sigfd > winch_sfd)
  {
    winch_nfds = sigfd + 1;
  }
  else
  {
    winch_nfds = winch_sfd + 1;
  }

  while (1)
  {
    // Flush the FSM's event queue
    rc = winch_fsm_engine();

    // Error in the FSM
    if (-1 == rc)
    {
      WINCH_ERR("FSM ERROR: '%m' (%d)\n", errno);
      rc = 1;
      goto err;
    }

    // End of FSM
    if (-2 == rc)
    {
      goto end;
    }

    // The previous event loop may have triggered the closing of file
    // descriptors. Each time they are closed, the fdset is updated. So,
    // the following select() is always passed an up to date fdset.

    // Wait for an event
    fdset = winch_fdset;

    // Used when the problems become tricky ;-)
    if (winch_dbg_level >= 8)
    {
    int *fds;
    int  nfds;

      rc = winch_fdset2table(&fdset, winch_nfds, &nfds, &fds);
      if (0 == rc)
      {
      int i;

	winch_display(2, "----> select(");
        for (i = 0; i < nfds; i ++)
	{
          winch_display(2, "%d%s", fds[i], (i < (nfds - 1) ? "," : ""));
	} // End for
	winch_display(2, ")\n");

        if (nfds)
	{
          free(fds);
	}
      }
    } // End if debug level

    rc = winch_select(winch_nfds, &fdset, 0, 0, 0);
    switch(rc)
    {
      case -1:
      {
        WINCH_ERR("select(): '%m' (%d)\n", errno);
        rc = 1;
        goto err;
      }
      break;

      default: // There is an event
      {
        ready = rc;

        // Signal ?
        if (ready && FD_ISSET(sigfd, &fdset))
        {
          ready --;

          rc = winch_read(sigfd, &siginfo, sizeof(siginfo));
          if (rc < 0)
	  {
            WINCH_ERR("read(): '%m' (%d)\n", errno);
            rc = 1;
            goto err;
	  }

          switch(siginfo.ssi_signo)
	  {
	    case SIGTERM:
	    case SIGINT:
	    {
              event = winch_new_event();
              if (!event)
	      {
                rc = 1;
                goto err;
	      }

              event->evt = WINCH_EVENT_SIGTERM;
              event->fsm = WINCH_FSM_SERVER;

              rc = winch_put_event(event);
	      if (rc != 0)
	      {
                winch_free_event(event);
                rc = 1;
                goto err;
	      }
            }
            break;

	    case SIGCHLD:
	    {
	    int status;
	    pid_t pid;

              // As signalfd() does provide multiple descriptors
              // if multiple occurrences of the same signal come,
              // we need to flush them
              while (1)
              {
                pid = waitpid(-1, &status, WNOHANG);
                if (pid <= 0)
                {
                  break;
                }

                event = winch_new_event();
                if (!event)
	        {
                  rc = 1;
                  goto err;
	        }

                event->evt = WINCH_EVENT_SIGCHLD;
                event->fsm = WINCH_FSM_CLIENT;
                event->client = winch_fsm_pid2client(pid);
                event->u.status = status;
                event->free_data = 0;

                rc = winch_put_event(event);
	        if (rc != 0)
	        {
                  winch_free_event(event);
                  rc = 1;
                  goto err;
	        }
	      } // End while
	    }
            break;

	    case SIGALRM:
	    {
	    winch_timer_t *timer;

              // Update the pending timers
	      winch_timer_update();

              // Get the elapsed timers
              do
	      {
                timer = winch_timer_elapsed();
                if (timer)
		{
                  event = (winch_event_t *)(timer->data);
                  winch_timer_free(timer);
                  rc = winch_put_event(event);
	          if (rc != 0)
	          {
                    winch_free_event(event);
                    rc = 1;
                    goto err;
	          }
		} // End if a timer elapsed
	      } while (timer);
	    }
            break;

	    default:
	    {
              WINCH_ERR("Unexpected signal %d\n", siginfo.ssi_signo);
              rc = 1;
              goto err;
	    }
	  } // End switch
	} // End if signal

        // Connection request
        if (ready && FD_ISSET(winch_sfd, &fdset))
        {
          ready --;

          event = winch_new_event();
          if (!event)
	  {
            rc = 1;
            goto err;
	  }

          event->evt = WINCH_EVENT_CNX;
          event->fsm = WINCH_FSM_SERVER;
          event->free_data = 0;

          rc = winch_put_event(event);
	  if (rc != 0)
	  {
            winch_free_event(event);
            rc = 1;
            goto err;
	  }
	} // End if connection request

        if (ready)
	{
          rc = winch_fsm_client_poll(&fdset);
          if (rc != 0)
	  {
            rc = 1;
            goto err;
	  }
	}
      }
      break;
    } // End switch
  } // End Main Loop

end:

  rc = 0;

err:

  if (winch_sfd >= 0)
  {
    shutdown(winch_sfd, SHUT_RDWR);
    close(winch_sfd);
  }

  if (sigfd >= 0)
  {
    close(sigfd);
  }

  return rc;

} // main
