// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_timer.h
// Description : Timer service (delta list principle)
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha           - Creation
//     08-Aug-2019  R. Koucha           - Enhanced the services (avoid
//                                        unecessary alarm() call)
//                                      - Some bug fixes
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#include "winch_link.h"
#include "winch_timer.h"
#include "winch_util.h"


// ----------------------------------------------------------------------------
// Name   : winch_timers
// Usage  : List of pending timers (delta list principle)
// ----------------------------------------------------------------------------
static winch_link_t winch_timers;


// ----------------------------------------------------------------------------
// Name   : winch_timer_run
// Usage  : Start the timer
// Return : None
// ----------------------------------------------------------------------------
static void winch_timer_run(void)
{
  (void)alarm(WINCH_TIMER_PERIOD);
} // winch_timer_run


// ----------------------------------------------------------------------------
// Name   : winch_timer_update
// Usage  : Decrement the pending timers (to be call at each alarm tick)
// Return : None
// ----------------------------------------------------------------------------
void winch_timer_update(void)
{
winch_timer_t *p;

  p = WINCH_LINK2TIMER(winch_timers.head);
  if (p)
  {
    assert(p->timeout > 0);
    p->timeout -= 1;
  }
} // winch_timer_update


// ----------------------------------------------------------------------------
// Name   : winch_timer_elapsed
// Usage  : Unlink the elpased timers. To be called until WINCH_TIMER_NULL is
//          returned.
//          The alarm is rearmed if there are still pending timers
// Return : Timer, if there is an elapsed timer
//          WINCH_TIMER_NULL, no more elapsed timers (alarm is rearmed if there
//          are still pending timers)
// ----------------------------------------------------------------------------
winch_timer_t *winch_timer_elapsed(void)
{
winch_timer_t *p;

  p = WINCH_LINK2TIMER(winch_timers.head);

  // If there are timers
  if (p)
  {
    // If timer elapsed
    if (0 == p->timeout)
    {
      winch_unlink(&(p->link), &winch_timers);

      WINCH_DBG(6, "Timer '%s' elapsed (%u s)\n", p->name, p->timer_val);

      return p;
    }
    else // The timer did not elapsed
    {
      // There are still pending timers ==> Rearm the alarm
      winch_timer_run();
    } // End if timer elapsed

  } // End if there are timers

  // No more timers, no need to rearm the alarm

  return WINCH_TIMER_NULL;
} // winch_timer_elapsed




// ----------------------------------------------------------------------------
// Name   : winch_timer_insert
// Usage  : Insert a timer in the delta list
// Return : Timer, if OK
//          WINCH_TIMER_NULL, if error
// ----------------------------------------------------------------------------
static void winch_timer_insert(winch_timer_t *timer)
{
winch_timer_t *p;

  // Insert the timer at the right place in the list
  p = WINCH_LINK2TIMER(winch_timers.head);

  // If the list of timers is empty
  if (!p)
  {
    // The timer is not running, start it
    winch_timer_run();
  }
  else // There are pending timers in the list
  {
    do
    {
      if (timer->timeout > p->timeout)
      {
        timer->timeout -= p->timeout;
      }
      else
      {
        // Insert the work before
        winch_link_before(&(timer->link), &(p->link), &winch_timers);

        p->timeout -= timer->timeout;
        return;
      }

      p = WINCH_LINK2TIMER(p->link.next);
    } while(p);

  } // End if the list of timers is empty

  // Either the list was empty or the timer goes at the end of the list
  winch_link_tail(&(timer->link), &winch_timers);

} // winch_timer_insert


// ----------------------------------------------------------------------------
// Name   : winch_timer_start
// Usage  : Start a new timer. If it is the sole timer in the list, the alarm
//          is rearmed.
// Return : Timer, if OK
//          WINCH_TIMER_NULL, if error
// ----------------------------------------------------------------------------
winch_timer_t *winch_timer_start(const char *name, unsigned int timeout, void *data)
{
winch_timer_t *timer;

  if (timeout < WINCH_TIMER_PERIOD)
  {
    errno = EINVAL;
    return WINCH_TIMER_NULL;
  }

  // Allocate a timer
  timer = (winch_timer_t *)malloc(sizeof(winch_timer_t));
  if (!timer)
  {
    // Errno is set
    return WINCH_TIMER_NULL;
  }

  // Populate the context
  timer->timeout = timeout;
  timer->timer_val = timeout;
  timer->name = name;
  timer->data = data;

  // Insert the timer at the right place in the list
  winch_timer_insert(timer);

  return timer;
} // winch_timer_start


// ----------------------------------------------------------------------------
// Name   : winch_timer_restart
// Usage  : Restart a running timer. If it is the sole timer in the list, the alarm
//          is rearmed.
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_timer_restart(winch_timer_t *timer, unsigned int timeout)
{
  if ((timeout < WINCH_TIMER_PERIOD) || !timer)
  {
    errno = EINVAL;
    return -1;
  }

  // Make sure that the timer is still linked
  if (!winch_is_linked(&(timer->link), &winch_timers))
  {
    errno = ENOENT;
    return -1;
  }

  // Add the time to the following timer if any
  if (timer->link.next)
  {
  winch_timer_t *timer_next;

    timer_next = WINCH_LINK2TIMER(timer->link.next);
    timer_next->timeout += timer->timeout;
  } // End if there is a following timer

  // Unlink the timer
  winch_unlink(&(timer->link), &winch_timers);

  // Populate the context (update the timeout)
  timer->timeout = timeout;
  timer->timer_val = timeout;

  // Reinsert the timer at the right place in the list
  winch_timer_insert(timer);

  return 0;
} // winch_timer_restart



// ----------------------------------------------------------------------------
// Name   : winch_timer_free
// Usage  : Free a timer
// Return : None
// ----------------------------------------------------------------------------
void winch_timer_free(winch_timer_t *timer)
{
  free(timer);
} // winch_timer_free


// ----------------------------------------------------------------------------
// Name   : winch_timer_stop
// Usage  : Stop a timer
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_timer_stop(winch_timer_t *timer)
{
  if (!timer)
  {
    errno = EINVAL;
    return -1;
  }

  // The timer may have elapsed
  if (!winch_is_linked(&(timer->link), &winch_timers))
  {
    errno = ENOENT;
    return -1;
  }

  // Add the time to the following timer if any
  if (timer->link.next)
  {
  winch_timer_t *timer_next;

    timer_next = WINCH_LINK2TIMER(timer->link.next);
    timer_next->timeout += timer->timeout;
  } // End if there is a following timer

  // Unlink the timer
  winch_unlink(&(timer->link), &winch_timers);

  // Stop the alarm if no more running timers
  if (winch_is_empty_list(&winch_timers))
  {
    alarm(0);
  }

  return 0;
} // winch_timer_stop


