// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_timer.h
// Description : Timer service
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha           - Creation
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=



#ifndef WINCH_TIMER_H
#define WINCH_TIMER_H

#include <stddef.h>

#include "winch_link.h"




// ----------------------------------------------------------------------------
// Name   : WINCH_TIMER_PERIOD
// Usage  : Timer's period in seconds
// ----------------------------------------------------------------------------
#define WINCH_TIMER_PERIOD 1


typedef struct
{
  const char   *name;
  unsigned int  timer_val;
  unsigned int  timeout;
  void         *data;
  winch_link_t  link;
} winch_timer_t;


// ----------------------------------------------------------------------------
// Name   : WINCH_LINK2TIMER
// Usage  : Get a timer from its link
// ----------------------------------------------------------------------------
#define WINCH_LINK2TIMER(p)  \
         (p ? ((winch_timer_t *)((char *)(p) - offsetof(winch_timer_t, link))) : NULL)


// ----------------------------------------------------------------------------
// Name   : WINCH_TIMER_NULL
// Usage  : NULL timer
// ----------------------------------------------------------------------------
#define WINCH_TIMER_NULL (winch_timer_t *)0


// ----------------------------------------------------------------------------
// Name   : winch_timer_start
// Usage  : Start a new timer. If it is the sole timer in the list, the alarm
//          is rearmed.
// Return : Timer, if OK
//          WINCH_TIMER_NULL, if error
// ----------------------------------------------------------------------------
extern winch_timer_t *winch_timer_start(const char *name, unsigned int timeout, void *data);


// ----------------------------------------------------------------------------
// Name   : winch_timer_stop
// Usage  : Stop a timer
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
extern int winch_timer_stop(winch_timer_t *timer);


// ----------------------------------------------------------------------------
// Name   : winch_timer_restart
// Usage  : Restart a running timer. If it is the sole timer in the list, the alarm
//          is rearmed.
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
extern int winch_timer_restart(winch_timer_t *timer, unsigned int timeout);


// ----------------------------------------------------------------------------
// Name   : winch_timer_update
// Usage  : Decrement the pending timers (to be call at each alarm tick)
// Return : None
// ----------------------------------------------------------------------------
extern void winch_timer_update(void);


// ----------------------------------------------------------------------------
// Name   : winch_timer_elapsed
// Usage  : Unlink the elpased timers. To be called until WINCH_TIMER_NULL is
//          returned.
//          The alarm is rearmed if there are still pending timers
// Return : Timer, if there is an elapsed timer
//          WINCH_TIMER_NULL, no more elapsed timers (alarm is rearmed if there
//          are still pending timers)
// ----------------------------------------------------------------------------
extern winch_timer_t *winch_timer_elapsed(void);


// ----------------------------------------------------------------------------
// Name   : winch_timer_free
// Usage  : Free a timer
// Return : None
// ----------------------------------------------------------------------------
extern void winch_timer_free(winch_timer_t *timer);




#endif // WINCH_TIMER_H

