// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_util.c
// Description : Utilities for winch programs
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha           - Creation
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>

#include "winch_util.h"


// ----------------------------------------------------------------------------
// Name   : winch_little_endian
// Usage  : Determine if the current host is little endian or not
// ----------------------------------------------------------------------------
static int winch_little_endian;


// ----------------------------------------------------------------------------
// Name   : winch_dbg_level
// Usage  : Debug level
// ----------------------------------------------------------------------------
unsigned int winch_dbg_level;


// ---------------------------------------------------------------------------
// Name  : winch_select
// Usage : select() system call with management of EINTR
// Return: Number of file descriptors contained in the 3 sets, if 0K
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
int winch_select(
                int             nfds,
                fd_set         *readfds,
                fd_set         *writefds,
                fd_set         *exceptfds,
                struct timeval *timeout
               )
{
int rc;

  do
  {
    rc = select(nfds, readfds, writefds, exceptfds, timeout);
  } while((rc < 0) && (EINTR == errno));

  return rc;
} // winch_select



// ---------------------------------------------------------------------------
// Name  : winch_ioctl
// Usage : ioctl() system call with management of EINTR
// Return: 0, if OK
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
int winch_ioctl(
                int             fd,
                int             cmd,
                void           *arg
               )
{
int rc;

  do
  {
    rc = ioctl(fd, cmd, arg);
  } while((rc < 0) && (EINTR == errno));

  return rc;
} // winch_ioctl



// ---------------------------------------------------------------------------
// Name  : winch_send
// Usage : send() system call with management of EINTR
// Return: Number of bytes sent, if OK
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
ssize_t winch_send(
                int             fd,
                const void     *buf,
                size_t          len,
                int             flags
               )
{
ssize_t rc;

  do
  {
    rc = send(fd, buf, len, flags);
  } while((rc < 0) && (EINTR == errno));

  return rc;
} // winch_send


// ---------------------------------------------------------------------------
// Name  : winch_recv
// Usage : recv() system call with management of EINTR
// Return: Number of bytes received
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
ssize_t winch_recv(
                int             fd,
                void           *buf,
                size_t          len,
                int             flags
		)
{
ssize_t rc;

  do
  {
    rc = recv(fd, buf, len, flags);
  } while((rc < 0) && (EINTR == errno));

  return rc;
} // winch_recv


// ---------------------------------------------------------------------------
// Name  : winch_read
// Usage : read() system call with management of EINTR
// Return: Number of read bytes, if 0K
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
ssize_t winch_read(
                  int     fd,
                  void   *buf,
                  size_t  count
                 )
{
ssize_t rc;

  do
  {
    rc = read(fd, buf, count);
  } while((rc < 0) && (EINTR == errno));

  return rc;
} // winch_read


// ---------------------------------------------------------------------------
// Name  : winch_write
// Usage : write() system call with management of EINTR
// Return: Number of written bytes, if 0K
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
ssize_t winch_write(
                  int         fd,
                  const void *buf,
                  size_t      count
                 )
{
ssize_t rc;

  do
  {
    rc = write(fd, buf, count);
  } while((rc < 0) && (EINTR == errno));

  return rc;
} // winch_write



// ---------------------------------------------------------------------------
// Name  : winch_display
// Usage : Write formatted data on a given file descriptor
// Return: None
// ---------------------------------------------------------------------------
void winch_display(
                   int fd,
                   const char *format,
                   ...
		  )
{
int     rc;
va_list ap;
char    buf[512];

  va_start(ap, format);
  rc = vsnprintf(buf, sizeof(buf), format, ap);
  va_end(ap);

  if (rc < 0)
  {
    // errno is set
    return;
  }

  if (rc >= (int)sizeof(buf))
  {
    errno = ENOSPC;
    return;
  }

  errno = 0;

  (void)winch_write(fd, buf, rc);

  // In case of error, errno is set

} // winch_display


//---------------------------------------------------------------------------
// Name  : winch_open_pty
// Usage : Allocate a PTY
// Note  : Cf. the online manual to understand the code below
//         (http://man7.org/linux/man-pages/man4/pts.4.html)
// Return: 0, if OK
//         -1, if error (errno is set)
//---------------------------------------------------------------------------
int winch_open_pty(
                  int    *pty_master,
                  int    *pty_slave,
                  char   *pty_name,
                  size_t  pty_name_sz
                 )
{
int     err_sav;
int     rc;

  if (!pty_master || !pty_slave || !pty_name || !pty_name_sz)
  {
    errno = EINVAL;
    return -1;
  }

  *pty_master = *pty_slave = -1;

  // Get a master pty
  //
  // posix_openpt() opens a pseudo-terminal master and returns its file
  // descriptor.
  // It is equivalent to open("/dev/ptmx",O_RDWR|O_NOCTTY) on Linux systems :
  //
  //       . O_RDWR Open the device for both reading and writing
  //       . O_NOCTTY Do not make this device the controlling terminal for the process
  *pty_master = posix_openpt(O_RDWR | O_NOCTTY);
  if (*pty_master < 0)
  {
    err_sav = errno;
    WINCH_ERR("Impossible to get a master pseudo-terminal - errno = '%m' (%d)\n", errno);
    goto error;
  }

  // Grant access to the slave pseudo-terminal
  // (Chown the slave to the calling user)
  if (0 != grantpt(*pty_master))
  {
    err_sav = errno;
    WINCH_ERR("Impossible to grant access to slave pseudo-terminal - errno = '%m' (%d)\n", errno);
    goto error;
  }

  // Unlock pseudo-terminal master/slave pair
  // (Release an internal lock so the slave can be opened)
  if (0 != unlockpt(*pty_master))
  {
    err_sav = errno;
    WINCH_ERR("Impossible to unlock pseudo-terminal master/slave pair - errno = '%m' (%d)\n", errno);
    goto error;
  }

  // Get the name of the slave pty
  rc = ptsname_r(*pty_master, pty_name, pty_name_sz);
  if (0 != rc)
  {
    err_sav = errno;
    WINCH_ERR("Impossible to get the name of the slave pseudo-terminal - errno = '%m' (%d)\n", errno);
    goto error;
  }

  // Open the slave part of the terminal
  *pty_slave = open(pty_name, O_RDWR);
  if (*pty_slave < 0)
  {
    err_sav = errno;
    WINCH_ERR("Impossible to open the slave pseudo-terminal (%s) - errno = '%m' (%d)\n", pty_name, errno);
    goto error;
  }

  return 0;

error:

  if (*pty_master >= 0)
  {
    close(*pty_master);
    *pty_master = -1;
  }

  if (*pty_slave >= 0)
  {
    close(*pty_slave);
    *pty_slave = -1;
  }

  pty_name[0] = '\0';

  errno = err_sav;

  return -1;

} // winch_open_pty


// ----------------------------------------------------------------------------
// Name   : winch_fdset2table
// Usage  : Return the set of file descriptors as a table of file descriptors 
// Return : 0, The table of file descriptors
//          (*nb_fds = number of entries may be 0)
//          -1, if error
// ----------------------------------------------------------------------------
int winch_fdset2table(
		       fd_set   *fdset,
                       int       nfds,
                       int      *nb_fds,
                       int     **fds     // Dynamically allocated
                     )
{
int i;

  // Get the number of file descriptors
  for (i = 0, (*nb_fds) = 0; i < nfds; i ++)
  {
    if (FD_ISSET(i, fdset))
    {
      (*nb_fds) ++;
    }
  } // End for

  // If there are file descriptors
  if (*nb_fds)
  {
  int j;

    *fds = (int *)malloc(*nb_fds * sizeof(int));
    if (!(*fds))
    {
      // errno is set
      return -1;
    }

    // Populate the table
    for (i = 0, j = 0; j < (*nb_fds); i ++)
    {
      if (FD_ISSET(i, fdset))
      {
        (*fds)[j ++] = i;
      }
    } // End for
  } // End if there are file descriptors in the set

  return 0;

} // winch_fdset2table


// ----------------------------------------------------------------------------
// Name   : winch_str2params
// Usage  : Return a parameter (i.e. ac, av[]) from a string
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
int winch_str2params(
		     char   *str,
                     int    *ac,
                     char ***av   // Dynamically allocated
                    )
{
char *p;
int   nb;

  if (!str)
  {
    errno = EINVAL;
    return -1;
  }

  if (!ac)
  {
    errno = EINVAL;
    return -1;
  }

  ac = 0;

  if (!av)
  {
    errno = EINVAL;
    return -1;
  }

  av = (char ***)0;
  nb = 1;

  *av = (char **)malloc(nb * sizeof(char *));
  if (!(*av))
  {
    // Errno  is set
    return -1;
  }
  (*av)[0] = (char *)0;

  p = str;
  while (*p)
  {
    while (isspace(*p))
    {
      p ++;
    } // End while

    if (*p)
    {
      // New word
      nb ++;
      *av = (char **)realloc(*av, nb * sizeof(char *));
      if (!(*av))
      {
        // Errno  is set
        return -1;
      }

      (*av)[nb - 2] = p;

      p += 1;
      while (*p && !isspace(*p))
      {
        p ++;
      } // End while

      // End of word
      if (*p)
      {
        *p ='\0';
        p ++;
      }
      else
      {
        // End of string
      }
    }
    else
    {
      // End of string
    }
  } // End while

  (*ac) = nb - 1;

  return 0;
} // winch_str2params


// ----------------------------------------------------------------------------
// Name   : winch_util_init
// Usage  : Initialize the utilities
// Return : None
// ----------------------------------------------------------------------------
void winch_util_init(void)
{
int            x = 1;
unsigned char *p;

  // Determine if we are big or little endian
  p = (unsigned char *)&x;
  winch_little_endian = ((*p) == (unsigned char)x);

  WINCH_DBG(2, "Host system is %s endian\n", (winch_little_endian ? "little" : "big"));

} // winch_util_init

