// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : winch_util.h
// Description : Utilities for winch programs
// License     :
//
//  Copyright (C) 2019 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     31-Jul-2019  R. Koucha           - Creation
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#ifndef WINCH_UTIL_H
#define WINCH_UTIL_H

#include <sys/types.h>
#include <sys/select.h>
#include <unistd.h>
#include <libgen.h>



// ----------------------------------------------------------------------------
// Name   : winch_dbg_level
// Usage  : Debug level
// ----------------------------------------------------------------------------
extern unsigned int winch_dbg_level;




// ---------------------------------------------------------------------------
// Name  : winch_display
// Usage : Write formatted data on a given file descriptor
// Return: None
// ---------------------------------------------------------------------------
extern void winch_display(
                   int fd,
                   const char *format,
                   ...
		   ) __attribute__ ((format (printf, 2, 3)));


// ----------------------------------------------------------------------------
// Name   : WINCH_ERR
// Usage  : Error messages
// ----------------------------------------------------------------------------
#define WINCH_ERR(format, ...) do {			       \
    winch_display(2,                                           \
            "WINCH ERROR(%d) - (%s/%s#%d): "format,            \
            getpid(),  basename(__FILE__), __FUNCTION__, __LINE__, ## __VA_ARGS__); \
                                 } while (0)


// ----------------------------------------------------------------------------
// Name   : WINCH_DISPLAY
// Usage  : Messages
// ----------------------------------------------------------------------------
#define WINCH_DISPLAY(format, ...) do {			       \
    winch_display(1,                                           \
            format,                                            \
            ## __VA_ARGS__);                                   \
                                 } while (0)

// ----------------------------------------------------------------------------
// Name   : WINCH_DBG
// Usage  : Debug messages
// ----------------------------------------------------------------------------
#define WINCH_DBG(level, format, ...) do {		       \
    if (level <= winch_dbg_level)                              \
      winch_display(2,                                           \
            "WINCH DBG(%d) - (%s/%s#%d): "format,            \
            level,  basename(__FILE__), __FUNCTION__, __LINE__, ## __VA_ARGS__); \
                                 } while (0)

// ---------------------------------------------------------------------------
// Name  : winch_select
// Usage : select() system call with management of EINTR
// Return: Number of file descriptors contained in the 3 sets, if 0K
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
extern int winch_select(
                int             nfds,
                fd_set         *readfds,
                fd_set         *writefds,
                fd_set         *exceptfds,
                struct timeval *timeout
		);


// ---------------------------------------------------------------------------
// Name  : winch_send
// Usage : send() system call with management of EINTR
// Return: Number of bytes sent
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
extern ssize_t winch_send(
                int             fd,
                const void     *buf,
                size_t          len,
                int             flags
		);


// ---------------------------------------------------------------------------
// Name  : winch_recv
// Usage : recv() system call with management of EINTR
// Return: Number of bytes received
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
extern ssize_t winch_recv(
                int             fd,
                void           *buf,
                size_t          len,
                int             flags
		);


// ---------------------------------------------------------------------------
// Name  : winch_ioctl
// Usage : ioctl() system call with management of EINTR
// Return: Number of file descriptors contained in the 3 sets, if 0K
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
extern int winch_ioctl(
                int             fd,
                int             cmd,
                void           *arg
		);


// ---------------------------------------------------------------------------
// Name  : winch_read
// Usage : read() system call with management of EINTR
// Return: Number of read bytes, if 0K
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
extern ssize_t winch_read(
                  int     fd,
                  void   *buf,
                  size_t  count
		  );


// ---------------------------------------------------------------------------
// Name  : winch_write
// Usage : write() system call with management of EINTR
// Return: Number of written bytes, if 0K
//         -1, if error (errno is set)
// ---------------------------------------------------------------------------
extern ssize_t winch_write(
                  int         fd,
                  const void *buf,
                  size_t      count
		  );



//---------------------------------------------------------------------------
// Name  : winch_open_pty
// Usage : Allocate a PTY
// Note  : Cf. the online manual to understand the code below
//         (http://man7.org/linux/man-pages/man4/pts.4.html)
// Return: 0, if OK
//         -1, if error (errno is set)
//---------------------------------------------------------------------------
extern int winch_open_pty(
                  int    *pty_master,
                  int    *pty_slave,
                  char   *pty_name,
                  size_t  pty_name_sz
		  );



// ----------------------------------------------------------------------------
// Name   : winch_fdset2table
// Usage  : Return the set of file descriptors as a table of file descriptors 
// Return : 0, The table of file descriptors
//          (*nb_fds = number of entries may be 0)
//          -1, if error
// ----------------------------------------------------------------------------
extern int winch_fdset2table(
		       fd_set   *fdset,
                       int       nfds,
                       int      *nb_fds,
                       int     **fds     // Dynamically allocated
                     );



// ----------------------------------------------------------------------------
// Name   : winch_str2params
// Usage  : Return a parameter (i.e. ac, av[]) from a string
// Return : 0, if OK
//          -1, if error
// ----------------------------------------------------------------------------
extern int winch_str2params(
		     char   *str,
                     int    *ac,
                     char ***av   // Dynamically allocated
		     );


// ----------------------------------------------------------------------------
// Name   : winch_util_init
// Usage  : Initialize the utilities
// Return : None
// ----------------------------------------------------------------------------
extern void winch_util_init(void);


#endif // WINCH_UTIL_H
